# Copyright(C) 2012-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections import OrderedDict
from functools import wraps
import importlib
import re
import base64
from hashlib import sha256
import zlib
from logging import Logger
from typing import Callable, Tuple, Type, Dict, Any, List, ClassVar

import os
from copy import copy, deepcopy
import inspect
from datetime import datetime, timedelta
from threading import Lock
from urllib.parse import urlparse, urljoin, urlencode, parse_qsl
import http
from uuid import uuid4
import warnings
import tempfile
import mimetypes

import urllib3
from dateutil import parser, tz
import requests

from woob.exceptions import (
    BrowserHTTPSDowngrade, BrowserRedirect, BrowserIncorrectPassword,
    BrowserUnavailable,
)
from woob.tools.date import now_as_utc
from woob.tools.log import getLogger
from woob.tools.json import json
from woob.tools.request import to_curl

from .adapters import HTTPAdapter
from .cookies import WoobCookieJar
from .exceptions import HTTPNotFound, ClientError, ServerError
from .har import HARManager
from .sessions import FuturesSession
from .profiles import Firefox, Profile
from .pages import NextPage
from .url import URL, normalize_url


class Browser:
    """
    Simple browser class.
    Acts like a browser, and doesn't try to do too much.

    >>> with Browser() as browser:
    ...     browser.open('https://example.org')
    ...
    <Response [200]>

    :param logger: parent logger (optional)
    :type logger: :py:class:`logging.Logger`
    :param proxy: use a proxy (dictionary with http/https as key and URI as value) (optional)
    :type proxy: dict
    :param responses_dirname: save responses to this directory (optional)
    :type responses_dirname: str
    :param proxy_headers: headers to supply to proxy (optional)
    :type proxy_headers: dict
    :param verify: either a boolean, in which case it controls whether we verify the server’s
        TLS certificate, or a string, in which case it must be a path to a CA bundle to use.
        Defaults will use the :attr:`Browser.VERIFY` attribute.
    :type verify: `None`, `bool` or `str`
    """

    PROFILE: ClassVar[Profile] = Firefox()
    """
    Default profile used by browser to navigate on websites.
    """

    TIMEOUT: ClassVar[float] = 10.0
    """
    Default timeout during requests.
    """

    REFRESH_MAX: ClassVar[float] = 0.0
    """
    When handling a Refresh header, the browsers considers it only if the sleep
    time in lesser than this value.
    """

    VERIFY: ClassVar[bool | str] = True
    """
    Check SSL certificates.

    If this is a string, path to the certificate or the CA bundle.

    Note that this value may be overriden by the ``verify`` argument on the
    constructor.
    """

    MAX_RETRIES: ClassVar[int] = 2
    """
    Maximum retries on failed requests.
    """

    MAX_WORKERS: ClassVar[int] = 10
    """
    Maximum of threads for asynchronous requests.
    """

    ALLOW_REFERRER: ClassVar[bool] = True
    """
    Controls how we send the ``Referer`` or not.

    If True, always allows the referers to be sent, False never, and None only
    if it is within the same domain.
    """

    HTTP_ADAPTER_CLASS: ClassVar[Type[HTTPAdapter]] = HTTPAdapter
    """
    Adapter class to use.
    """

    COOKIE_POLICY: ClassVar[http.cookiejar.CookiePolicy | None] = None
    """
    Default CookieJar policy.

    Example: :class:`~woob.browser.cookies.BlockAllCookies()`
    """

    @classmethod
    def asset(cls, localfile: str) -> str:
        """
        Absolute file path for a module local file.
        """
        if os.path.isabs(localfile):
            return localfile
        return os.path.join(os.path.dirname(inspect.getfile(cls)), localfile)

    def __new__(cls, *args, **kwargs):
        """ Accept any arguments, necessary for AbstractBrowser __new__ override.

        AbstractBrowser, in its overridden __new__, removes itself from class hierarchy
        so its __new__ is called only once. In python 3, default (object) __new__ is
        then used for next instantiations but it's a slot/"fixed" version supporting
        only one argument (type to instanciate).
        """
        return object.__new__(cls)

    def __init__(
        self,
        logger: Logger | None = None,
        proxy: Dict[str, str] | None = None,
        responses_dirname: str | None = None,
        proxy_headers: Dict[str, str] | None = None,
        woob: None = None,
        weboob: None = None,
        *,
        verify: bool | str | None = None,
    ):

        if woob is not None or weboob is not None:
            warnings.warn(
                "Don't use the 'woob' and 'weboob' parameters, they will be removed in woob 4.0",
                DeprecationWarning,
                stacklevel=2,
            )

        if logger:
            self.logger = getLogger("browser", logger)
        else:
            self.logger = getLogger("woob.browser")

        self.responses_dirname = responses_dirname
        self.responses_count = 0
        self.responses_lock = Lock()

        if self.logger.settings['ssl_insecure']:
            self.verify = False
        elif verify is not None:
            self.verify = verify
        else:
            self.verify = self.VERIFY

        if isinstance(self.verify, str):
            self.verify = self.asset(self.verify)

        self.PROXIES = proxy or {}
        self.proxy_headers = proxy_headers or {}
        self._setup_session(self.PROFILE)
        self.url: str | None = None
        self.response: requests.Response | None = None
        self.har_manager: HARManager | None = None

        if self.responses_dirname is not None:
            self.har_manager = HARManager(self.responses_dirname, self.logger)

    def deinit(self):
        """
        Deinitialisation of the browser.

        Call it when you stop to use the browser and you don't use it in a
        context manager.

        Can be overrided by any subclass which wants to cleanup after browser
        usage.
        """
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.deinit()

    def set_normalized_url(self, response: requests.Response, **kwargs):
        """
        Set the normalized URL on the response.

        :param response: the response to change
        :type response: :class:`requests.Response`
        """
        response.url = normalize_url(response.url)

    def save_response(self, response: requests.Response, warning: bool = False, **kwargs):
        """
        Save responses.

        By default it creates an HAR file and append request and response in.

        If ``WOOB_USE_OBSOLETE_RESPONSES_DIR`` is set to 1, it'll create a
        directory and all requests will be saved in three files:

        * ``0X-url-request.txt``
        * ``0X-url-response.txt``
        * ``0X-url.EXT``

        Information about which files are created is display in logs.

        Also if ``WOOB_CURLIFY_REQUEST`` is set to 1, ``0X-url-request.txt``
        will be filled with a ready to use curl command based on the request.

        :param response: the response to save
        :type response: :class:`requests.Response`
        :param warning: if True, display the saving logs as warnings (default to False)
        :type warning: bool
        """
        if self.responses_dirname is None:
            self.responses_dirname = tempfile.mkdtemp(prefix='woob_session_')

            self.logger.info('Debug data will be saved in this directory: %s', self.responses_dirname)
        elif not os.path.isdir(self.responses_dirname):
            os.makedirs(self.responses_dirname)

        if self.har_manager is None:
            self.har_manager = HARManager(self.responses_dirname, self.logger)

        slug = uuid4().hex

        with self.responses_lock:
            counter = self.responses_count
            self.responses_count += 1

        response_filepath = slug

        if os.environ.get('WOOB_USE_OBSOLETE_RESPONSES_DIR') == '1':
            # get the content-type, remove optionnal charset part
            mimetype = response.headers.get('Content-Type', '').split(';')[0]

            # try to get an extension (and avoid adding 'None')
            ext = mimetypes.guess_extension(mimetype, False) or ''

            filename = '%02d-%d-%s%s' % \
                (counter, response.status_code, slug, ext)

            response_filepath = os.path.join(self.responses_dirname, filename)

            request = response.request
            with open(response_filepath + '-request.txt', 'w', encoding='utf-8') as f:
                f.write('%s %s\n\n\n' % (request.method, request.url))

                for key, value in request.headers.items():
                    f.write('%s: %s\n' % (key, value))
                if request.body is not None:  # separate '' from None
                    body = request.body if isinstance(request.body, str) else request.body.decode()
                    f.write('\n\n\n%s' % body)
                if os.environ.get('WOOB_CURLIFY_REQUEST') == '1':
                    curl = to_curl(request)
                    f.write('\n\n' + curl + '\n')
            with open(response_filepath + '-response.txt', 'w', encoding='utf-8') as f:
                if hasattr(response.elapsed, 'total_seconds'):
                    f.write('Time: %3.3fs\n' % response.elapsed.total_seconds())
                f.write('%s %s\n\n\n' % (response.status_code, response.reason))
                for key, value in response.headers.items():
                    f.write('%s: %s\n' % (key, value))

            with open(response_filepath, 'wb') as f:
                f.write(response.content)

            match_filepath = os.path.join(self.responses_dirname, 'url_response_match.txt')
            with open(match_filepath, 'a', encoding='utf-8') as f:
                f.write('# %d %s %s\n' % (response.status_code, response.reason, response.headers.get('Content-Type', '')))
                f.write('%s\t%s\n' % (response.url, filename))

        self.har_manager.save_response(slug, response)

        msg = 'Response saved to %s'
        if warning:
            self.logger.warning(msg, response_filepath)
        else:
            self.logger.info(msg, response_filepath)

    def _save_request_to_har(self, request):
        assert self.har_manager is not None
        assert self.responses_dirname is not None

        # called when we don't have any response object
        if not os.path.isdir(self.responses_dirname):
            os.makedirs(self.responses_dirname)

        request_filepath = slug = uuid4().hex

        with self.responses_lock:
            counter = self.responses_count
            self.responses_count += 1

        time = self.TIMEOUT * 1000  # because TIMEOUT is in seconds, and we want milliseconds
        self.har_manager.save_request_only(slug, request, time)

        if os.environ.get('WOOB_USE_OBSOLETE_RESPONSES_DIR') == '1':
            request_filepath = os.path.join(
                self.responses_dirname,
                '%02d-000-%s-request.txt' % (counter, slug),
            )

            with open(request_filepath, 'w', encoding='utf-8') as f:
                f.write('%s %s\n\n\n' % (request.method, request.url))

                for key, value in request.headers.items():
                    f.write('%s: %s\n' % (key, value))

                if request.body is not None:  # separate '' from None
                    body = (
                        request.body if isinstance(request.body, str)
                        else request.body.decode()
                    )
                    f.write('\n\n\n%s' % body)

                if os.environ.get('WOOB_CURLIFY_REQUEST') == '1':
                    curl = to_curl(request)
                    f.write('\n\n' + curl + '\n')

        self.logger.info('Request saved to %s', request_filepath)

    def _create_session(self) -> requests.Session:
        return FuturesSession(
            max_workers=self.MAX_WORKERS, max_retries=self.MAX_RETRIES,
            adapter_class=self.HTTP_ADAPTER_CLASS,
        )

    def _setup_session(self, profile: Profile):
        """
        Set up a python3-requests session for our usage.
        """
        session = self._create_session()

        session.proxies = self.PROXIES

        session.verify = self.verify

        if not session.verify:
            try:
                urllib3.disable_warnings()
            except AttributeError:
                # urllib3 is too old, warnings won't be disable
                pass

        adapter_kwargs: Dict[str, Any] = {}

        # defines a max_retries. It's mandatory in case a server is not
        # handling keep alive correctly, like the proxy burp
        adapter_kwargs['max_retries'] = self.MAX_RETRIES

        adapter_kwargs['proxy_headers'] = self.proxy_headers

        # set connection pool size equal to MAX_WORKERS if needed
        if self.MAX_WORKERS > requests.adapters.DEFAULT_POOLSIZE:
            adapter_kwargs['pool_connections'] = self.MAX_WORKERS
            adapter_kwargs['pool_maxsize'] = self.MAX_WORKERS

        session.mount('http://', self.HTTP_ADAPTER_CLASS(**adapter_kwargs))
        session.mount('https://', self.HTTP_ADAPTER_CLASS(**adapter_kwargs))

        ## woob only can provide proxy and HTTP auth options
        session.trust_env = False

        profile.setup_session(session)

        session.hooks['response'].append(self.set_normalized_url)
        if self.responses_dirname is not None:
            session.hooks['response'].append(self.save_response)

        self.session = session

        session.cookies = WoobCookieJar()
        if self.COOKIE_POLICY:
            session.cookies.set_policy(self.COOKIE_POLICY)

    def set_profile(self, profile: Profile):
        """
        Update the profile of the session.
        """
        profile.setup_session(self.session)

    def location(self, url: str | requests.Request, **kwargs) -> requests.Response:
        """
        Like :meth:`open()` but also changes the current URL and response.
        This is the most common method to request web pages.

        Other than that, has the exact same behavior of :meth:`open()`.
        """
        assert not kwargs.get('is_async'), "Please use open() instead of location() to make asynchronous requests."
        response = self.open(url, **kwargs)
        self.response = response
        self.url = self.response.url
        return response

    def open(
        self,
        url: str | requests.Request,
        *,
        referrer: str | None = None,
        allow_redirects: bool = True,
        stream: bool | None = None,
        timeout: float | None = None,
        verify: str | bool | None = None,
        cert: str | Tuple[str, str] | None = None,
        proxies: Dict | None = None,
        data_encoding: str | None = None,
        is_async: bool = False,
        callback: Callable[[requests.Response], requests.Response] | None = None,
        **kwargs
    ) -> requests.Response:
        """
        Make an HTTP request like a browser does:
         * follow redirects (unless disabled)
         * provide referrers (unless disabled)

        Unless a ``method`` is explicitly provided, it makes a GET request,
        or a POST if data is not None,
        An empty ``data`` (like ``''`` or ``{}``, not ``None``) *will* make a POST.

        It is a wrapper around session.request().
        All ``session.request()`` options are available.
        You should use :meth:`location()` or :meth:`open()` and not ``session.request()``,
        since it has some interesting additions, which are easily individually
        disabled through the arguments.

        Call this instead of :meth:`location()` if you do not want to "visit" the URL
        (for instance, you are downloading a file).

        When ``is_async`` is ``True``, :meth:`open()` returns a :py:class:`~concurrent.futures.Future` object (see
        :py:mod:`concurrent.futures` for more details), which can be evaluated with its
        :py:meth:`~concurrent.futures.Future.result()` method. If any exception is raised while processing request,
        it is caught and re-raised when calling :py:meth:`~concurrent.futures.Future.result()`.

        For example:

        >>> Browser().open('https://google.com', is_async=True).result().text # doctest: +SKIP

        :param url: URL
        :param params: (optional) Dictionary, list of tuples or bytes to send
            in the query string
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body
        :param json: (optional) A JSON serializable Python object to send in the body
        :param headers: (optional) Dictionary of HTTP Headers to send
        :param cookies: (optional) Dict or CookieJar object to send
        :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
            ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
            or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
            defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
            to add for the file.
        :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.

        :param referrer: (optional) Force referrer. False to disable sending it, None for guessing
        :type referrer: str or False or None

        :param allow_redirects: (optional) if ``True``, follow HTTP redirects (default: ``True``)
        :type allow_redirects: bool

        :param stream: (optional) if ``False``, the response content will be immediately downloaded.
        :param timeout: (optional) How many seconds to wait for the server to send data
                        before giving up, as a float, or a tuple.
        :type timeout: float or tuple
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
                the server's TLS certificate, or a string, in which case it must be a path
                to a CA bundle to use. If not provided, uses the :attr:`Browser.VERIFY` class attribute value, the
                :attr:`Browser.verify` attribute one, or ``True``.
        :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
        :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.

        :param is_async: (optional) Process request in a non-blocking way (default: ``False``)
        :type is_async: bool

        :param callback: (optional) Callback to be called when request has finished,
                         with response as its first and only argument
        :type callback: callable

        :return: :class:`requests.Response <Response>` object
        :rtype: :class:`requests.Response`
        """

        if isinstance(url, str):
            url = normalize_url(url)
        elif isinstance(url, requests.Request):
            url.url = normalize_url(url.url)

        req = self.build_request(url, referrer=referrer, data_encoding=data_encoding, **kwargs)
        preq = self.prepare_request(req)

        if hasattr(preq, '_cookies'):
            # The _cookies attribute is not present in requests < 2.2. As in
            # previous version it doesn't calls extract_cookies_to_jar(), it is
            # not a problem as we keep our own cookiejar instance.
            preq._cookies = WoobCookieJar.from_cookiejar(preq._cookies)
            if self.COOKIE_POLICY:
                preq._cookies.set_policy(self.COOKIE_POLICY)

        if proxies is None:
            proxies = self.PROXIES

        if verify is None:
            verify = self.verify

        if timeout is None:
            timeout = self.TIMEOUT
        if callback is None:
            callback = lambda response: response

        # We define an inner_callback here in order to execute the same code
        # regardless of is_async param.
        def inner_callback(future, response):
            if allow_redirects:
                response = self.handle_refresh(response)

            self.raise_for_status(response)
            return callback(response)

        # call python3-requests
        try:
            response = self.session.send(preq,
                                         allow_redirects=allow_redirects,
                                         stream=stream,
                                         timeout=timeout,
                                         verify=verify,
                                         cert=cert,
                                         proxies=proxies,
                                         callback=inner_callback,
                                         is_async=is_async)
        except Exception as error:
            # response in these kind of exception are already stored in HAR
            # skip them to not store response twice
            already_handled_exception = any(
                isinstance(error, exc) for exc in (HTTPNotFound, ClientError, ServerError)
            )
            if not already_handled_exception and self.responses_dirname is not None and self.har_manager is not None:
                # when timeout or any kind of exception occur
                # we don't have any response given by python-requests
                # but we still need to store request to HAR
                if isinstance(error, requests.exceptions.RequestException):
                    if error.response is not None:
                        slug = uuid4().hex
                        self.har_manager.save_response(slug, error.response)
                    elif error.request:
                        # some exceptions raised by python-requests don't even provide request
                        # we prefer use error.request because python3-requests may modify request
                        # by calling add_headers() to it, but it won't happens if
                        # requests.adapters.HTTPAdapter.add_headers() is not overwritten
                        self._save_request_to_har(error.request)
                    else:
                        # use preq request if the exception doesn't provide the request
                        self._save_request_to_har(preq)
                # don't store request for other exceptions, it can be anything
                # but in that case request has already been stored in HAR file
            raise
        return response

    def async_open(self, url: str, **kwargs) -> requests.Response:
        """
        Shortcut to open(url, is_async=True).
        """
        if 'async' in kwargs:
            del kwargs['async']
        if 'is_async' in kwargs:
            del kwargs['is_async']
        return self.open(url, is_async=True, **kwargs)

    def raise_for_status(self, response: requests.Response):
        """
        Like :meth:`requests.Response.raise_for_status()` but will use other
        exception specific classes:

        * :class:`~woob.browser.exceptions.HTTPNotFound` for 404
        * :class:`~woob.browser.exceptions.ClientError` for 4xx errors
        * :class:`~woob.browser.exceptions.ServerError` for 5xx errors
        """
        if 400 <= response.status_code < 500:
            http_error_msg = '%s Client Error: %s' % (response.status_code, response.reason)
            if response.status_code == 404:
                raise HTTPNotFound(http_error_msg, response=response)
            raise ClientError(http_error_msg, response=response)
        elif 500 <= response.status_code < 600:
            http_error_msg = '%s Server Error: %s' % (response.status_code, response.reason)
            raise ServerError(http_error_msg, response=response)

        # in case we did not catch something that should be
        response.raise_for_status()

    def build_request(
        self,
        url: str | requests.Request,
        *,
        referrer: str | None = None,
        data_encoding: str | None = None,
        **kwargs
    ) -> requests.Request:
        """
        Does the same job as :meth:`open()`, but returns a :class:`~requests.Request` without
        submitting it.
        This allows further customization to the :class:`~requests.Request`.
        """

        url_string: str
        if isinstance(url, requests.Request):
            req = url
            url_string = req.url
        elif isinstance(url, str):
            req = requests.Request(url=url, **kwargs)
            url_string = url
        else:
            raise TypeError('"url" must be a string or a requests.Request object.')

        # guess method
        if req.method is None:
            # 'data' and 'json' (even if empty) are (always?) passed to build_request
            # and None is their default. For a Request object, the defaults are different.
            # Request.json is None and Request.data == [] by default.
            # Could they break unexpectedly?
            if (
                req.data or kwargs.get('data') is not None
                or req.json or kwargs.get('json') is not None
            ):
                req.method = 'POST'
            else:
                req.method = 'GET'

        # convert unicode strings to proper encoding
        if isinstance(req.data, str) and data_encoding:
            req.data = req.data.encode(data_encoding)
        if isinstance(req.data, dict) and data_encoding:
            encoded_data = OrderedDict()
            for k, v in req.data.items():
                encoded_value: Any
                if isinstance(v, str):
                    encoded_value = v.encode(data_encoding)
                elif isinstance(v, list):
                    encoded_value = [
                        element.encode(data_encoding) if isinstance(element, str) else element for element in v
                    ]
                else:
                    encoded_value = v
                encoded_data[k] = encoded_value
            req.data = encoded_data

        if referrer is None:
            referrer = self.get_referrer(self.url, url_string)
        if referrer:
            # Yes, it is a misspelling.
            req.headers.setdefault('Referer', referrer)

        return req

    def prepare_request(self, req: requests.Request) -> requests.PreparedRequest:
        """
        Get a prepared request from a :class:`~requests.Request` object.

        This method aims to be overloaded by children classes.
        """
        return self.session.prepare_request(req)

    REFRESH_RE = re.compile(r"^(?P<sleep>[\d\.]+)(;\s*url=[\"']?(?P<url>.*?)[\"']?)?$", re.IGNORECASE)

    def handle_refresh(self, response: requests.Response) -> requests.Response:
        """
        Called by open, to handle Refresh HTTP header.

        It only redirect to the refresh URL if the sleep time is inferior to
        :attr:`REFRESH_MAX`.
        """
        if 'Refresh' not in response.headers:
            return response

        m = self.REFRESH_RE.match(response.headers['Refresh'])
        if m:
            # XXX perhaps we should not redirect if the refresh url is equal to the current url.
            url = m.groupdict().get('url', None) or response.request.url
            sleep = float(m.groupdict()['sleep'])

            assert isinstance(url, str)

            if sleep <= self.REFRESH_MAX:
                self.logger.debug('Refresh to %s', url)
                return self.open(url)
            else:
                self.logger.debug('Do not refresh to %s because %s > REFRESH_MAX(%s)', url, sleep, self.REFRESH_MAX)
                return response

        self.logger.warning('Unable to handle refresh "%s"', response.headers['Refresh'])

        return response

    def get_referrer(self, oldurl: str | None, newurl: str) -> str | None:
        """
        Get the referrer to send when doing a request.
        If we should not send a referrer, it will return None.

        Reference: https://en.wikipedia.org/wiki/HTTP_referer

        The behavior can be controlled through the ALLOW_REFERRER attribute.
        True always allows the referers
        to be sent, False never, and None only if it is within
        the same domain.

        :param oldurl: Current absolute URL
        :type oldurl: str or None

        :param newurl: Target absolute URL
        :type newurl: str

        :rtype: str or None
        """
        if self.ALLOW_REFERRER is False:
            return None
        if oldurl is None:
            return None
        old = urlparse(oldurl)
        new = urlparse(newurl)
        # Do not leak secure URLs to insecure URLs
        if old.scheme == 'https' and new.scheme != 'https':
            return None
        # Reloading the page. Usually no referrer.
        if oldurl == newurl:
            return None
        # Domain-based privacy
        if self.ALLOW_REFERRER is None and old.netloc != new.netloc:
            return None
        return oldurl

    def export_session(self) -> dict:
        """
        Export session into a dict.

        Default format is::

            {
                'url': last_url,
                'cookies': cookies_dict
            }

        You should store it as is.
        """
        # XXX similar to StatesMixin, should be merged?
        def make_cookie(c):
            d = {
                k: getattr(c, k) for k in ['name', 'value', 'domain', 'path', 'secure']
            }
            #d['session'] = c.discard
            d['httpOnly'] = 'httponly' in [k.lower() for k in c._rest.keys()]
            d['expirationDate'] = getattr(c, 'expires', None)
            return d

        return {
            'url': self.url,
            'cookies': [make_cookie(c) for c in self.session.cookies],
        }


class UrlNotAllowed(Exception):
    """
    Raises by :class:`DomainBrowser` when :attr:`~DomainBrowser.RESTRICT_URL` is set and trying to go
    on an url not matching :attr:`~DomainBrowser.BASEURL`.
    """


class DomainBrowser(Browser):
    """
    A browser that handles relative URLs and can have a base URL (usually a domain).

    For instance ``self.location('/hello')`` will get https://woob.tech/hello
    if :attr:`BASEURL` is ``'https://woob.tech/'``.

    >>> class ExampleBrowser(DomainBrowser):
    ...     BASEURL = 'https://example.org'
    ...
    >>> with ExampleBrowser() as browser:
    ...     browser.open('/')
    ...
    <Response [200]>
    """

    BASEURL: str | None = None
    """
    Base URL, e.g. ``'https://woob.tech/'``.

    See :meth:`absurl()`.
    """

    RESTRICT_URL: ClassVar[bool | List[str]] = False
    """
    URLs allowed to load.
    This can be used to force SSL (if the :attr:`BASEURL` is SSL) or any other leakage.
    Set to ``True`` to allow only URLs starting by the :attr:`BASEURL`.
    Set it to a list of allowed URLs if you have multiple allowed URLs.
    More complex behavior is possible by overloading :meth:`url_allowed()`.
    """

    def __init__(
        self,
        baseurl: str | None = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        if baseurl is not None:
            self.BASEURL = baseurl

    def url_allowed(self, url: str) -> bool:
        """
        Checks if we are allowed to visit an URL.
        See RESTRICT_URL.

        :param url: Absolute URL
        :type url: str
        :rtype: bool
        """
        if self.BASEURL is None or self.RESTRICT_URL is False:
            return True
        if self.RESTRICT_URL is True:
            return url.startswith(self.BASEURL)
        for restrict_url in self.RESTRICT_URL:
            if url.startswith(restrict_url):
                return True
        return False

    def absurl(self, uri: str, base: str | bool | None = None) -> str:
        """
        Get the absolute URL, relative to a base URL.
        If base is ``None``, it will try to use the current URL.
        If there is no current URL, it will try to use :attr:`BASEURL`.

        If base is ``False``, it will always try to use the current URL.
        If base is ``True``, it will always try to use BASEURL.

        :param uri: URI to make absolute. It can be already absolute.
        :type uri: str

        :param base: Base absolute URL.
        :type base: str or None or False or True

        :rtype: str
        """
        if not base:
            base = self.url
        if base is None or base is True:
            base = self.BASEURL

        return urljoin(base, uri)

    def open(self, url: requests.Request | str, *args, **kwargs) -> requests.Response:
        """
        Like :meth:`Browser.open` but handles urls without domains, using
        the :attr:`BASEURL` attribute.
        """
        if isinstance(url, requests.Request):
            req = url
            req_url = req.url
        else:
            req = None
            req_url = url

        abs_url = self.absurl(req_url)
        if not self.url_allowed(abs_url):
            raise UrlNotAllowed(abs_url)

        if req:
            req.url = abs_url
            url = req
        else:
            url = abs_url
        return super().open(url, *args, **kwargs)

    def go_home(self) -> requests.Response:
        """
        Go to the "home" page, usually the BASEURL.
        """
        return self.location(self.BASEURL or self.absurl('/'))


class PagesBrowser(DomainBrowser):
    r"""
    A browser which works pages and keep state of navigation.

    To use it, you have to derive it and to create :class:`~woob.browser.url.URL` objects as class attributes. When
    :meth:`open()` or :meth:`location()` are called, if the url matches one of :class:`~woob.browser.url.URL` objects, it returns a
    :class:`~woob.browser.pages.Page` object. In case of :meth:`location()`, it stores it in ``self.page``.

    Example:

        >>> import re
        >>> from .pages import HTMLPage
        >>> class ListPage(HTMLPage):
        ...     def get_items(self):
        ...         for link in self.doc.xpath('//a[matches(@href, "list-\d+.html")]/@href'):
        ...             yield re.match('list-(\d+).html', link).group(1)
        ...
        >>> class ItemPage(HTMLPage):
        ...     def iter_values(self):
        ...         for el in self.doc.xpath('//li'):
        ...             yield el.text
        ...
        >>> class MyBrowser(PagesBrowser):
        ...     BASEURL = 'https://woob.tech/tests/'
        ...     list = URL(r'$', ListPage)
        ...     item = URL(r'list-(?P<id>\d+)\.html', ItemPage)
        ...
        >>> b = MyBrowser()
        >>> b.list.go()
        <woob.browser.browsers.ListPage object at 0x...>
        >>> b.page.url
        'https://woob.tech/tests/'
        >>> list(b.page.get_items())
        ['1', '2']
        >>> b.item.build(id=42)
        'https://woob.tech/tests/list-42.html'
        >>> b.item.go(id=1)
        <woob.browser.browsers.ItemPage object at 0x...>
        >>> list(b.page.iter_values())
        ['One', 'Two']
    """

    _urls = None

    def __init__(self, *args, **kwargs):
        self._urls = OrderedDict()
        self.highlight_el = kwargs.pop('highlight_el', False)
        super().__init__(*args, **kwargs)

        self.page = None

        # exclude properties because they can access other fields not yet defined
        def is_property(attr):
            v = getattr(type(self), attr, None)
            return hasattr(v, '__get__') or hasattr(v, '__set__')

        attrs = [(attr, getattr(self, attr)) for attr in dir(self) if not is_property(attr)]
        attrs = [v for v in attrs if isinstance(v[1], URL)]
        attrs.sort(key=lambda v: v[1]._creation_counter)
        for k, v in deepcopy(attrs):
            self._urls[k] = v
            setattr(self, k, v)
        for url in self._urls.values():
            url.browser = self

    def __setattr__(self, key, value):
        if isinstance(self._urls, OrderedDict):
            # _urls is instanciated, we can now feed it accordingly.
            if isinstance(value, URL):
                # We want to either replace in-place, or add to the URLs.
                if key in self._urls:
                    # We want to actually make the old URL unusable.
                    self._urls[key].browser = None

                value = copy(value)
                value.browser = self
                self._urls[key] = value
            elif key in self._urls:
                # We want to remove the URL from our mapping only.
                url = self._urls.pop(key)
                url.browser = None

        super().__setattr__(key, value)

    def __delattr__(self, key):
        if isinstance(self._urls, OrderedDict):
            if key in self._urls:
                # We want to remove the URL from our mapping.
                del self._urls[key]

        super().__delattr__(key)

    def open(self, *args, **kwargs) -> requests.Response:
        """
        Same method than
        :meth:`~woob.browser.browsers.DomainBrowser.open`, but the
        response contains an attribute ``page`` if the url matches any
        :class:`~woob.browser.url.URL` object.
        """

        callback = kwargs.pop('callback', lambda response: response)
        page_class = kwargs.pop('page', None)

        # Have to define a callback to seamlessly process synchronous and
        # asynchronous requests, see :meth:`Browser.open` and its `is_async`
        # and `callback` params.
        def internal_callback(response):
            # Try to handle the response page with an URL instance.
            response.page = None
            if page_class:
                response.page = page_class(self, response)
                return callback(response)

            for url in self._urls.values():
                response.page = url.handle(response)
                if response.page is not None:
                    self.logger.debug('Handle %s with %s', response.url, response.page.__class__.__name__)
                    break

            if response.page is None:
                regexp = r'^(?P<proto>\w+)://.*'

                proto_response = re.match(regexp, response.url)
                if proto_response and self.BASEURL:
                    proto_response = proto_response.group('proto')
                    proto_base = re.match(regexp, self.BASEURL).group('proto')

                    if proto_base == 'https' and proto_response != 'https':
                        raise BrowserHTTPSDowngrade()

                self.logger.debug('Unable to handle %s', response.url)

            return callback(response)

        return super(PagesBrowser, self).open(callback=internal_callback, *args, **kwargs)

    def location(self, *args, **kwargs) -> requests.Response:
        """
        Same method than :meth:`~woob.browser.browsers.Browser.location`, but
        if the url matches any :class:`~woob.browser.url.URL` object, an
        attribute ``page`` is added to response, and the attribute :attr:`page`
        is set on the browser.
        """
        if self.page is not None:
            # Call leave hook.
            self.page.on_leave()

        response = self.open(*args, **kwargs)

        self.response = response
        self.page = response.page
        self.url = response.url

        if self.page is not None:
            # Call load hook.
            self.page.on_load()

        # Returns self.response in case on_load recalls location()
        return self.response

    def pagination(self, func: Callable, *args, **kwargs):
        r"""
        This helper function can be used to handle pagination pages easily.

        When the called function raises an exception :class:`~woob.browser.pages.NextPage`, it goes
        on the wanted page and recall the function.

        :class:`~woob.browser.pages.NextPage` constructor can take an url or a Request object.

        >>> from .pages import HTMLPage
        >>> class Page(HTMLPage):
        ...     def iter_values(self):
        ...         for el in self.doc.xpath('//li'):
        ...             yield el.text
        ...         for next in self.doc.xpath('//a'):
        ...             raise NextPage(next.attrib['href'])
        ...
        >>> class Browser(PagesBrowser):
        ...     BASEURL = 'https://woob.tech'
        ...     list = URL('/tests/list-(?P<pagenum>\d+).html', Page)
        ...
        >>> b = Browser()
        >>> b.list.go(pagenum=1) # doctest: +ELLIPSIS
        <woob.browser.browsers.Page object at 0x...>
        >>> list(b.pagination(lambda: b.page.iter_values()))
        ['One', 'Two', 'Three', 'Four']

        .. note: consider using :func:`~woob.browser.pages.pagination` decorator instead.
        """
        while True:
            try:
                for r in func(*args, **kwargs):
                    yield r
            except NextPage as e:
                self.location(e.request)
            else:
                return


def need_login(func):
    """
    Decorator used to require to be logged to access to this function.

    This decorator can be used on any method whose first argument is a
    browser (typically a :class:`LoginBrowser`). It checks for the
    ``logged`` attribute in the current browser's page: when this
    attribute is set to ``True`` (e.g., when the page inherits
    :class:`~woob.browser.pages.LoggedPage`), then nothing special happens.

    In all other cases (when the browser isn't on any defined page or
    when the page's ``logged`` attribute is ``False``), the
    :meth:`LoginBrowser.do_login` method of the browser is called before
    calling :`func`.
    """

    @wraps(func)
    def inner(browser: LoginBrowser, *args, **kwargs):
        if (
            (
                not hasattr(browser, 'logged') or
                (
                    hasattr(browser, 'logged') and
                    not browser.logged
                )
            ) and (
                not hasattr(browser, 'page') or
                browser.page is None or
                not browser.page.logged
            )
        ):
            browser.do_login()
            if browser.logger.settings.get('export_session'):
                browser.logger.debug('logged in with session: %s', json.dumps(browser.export_session()))
        return func(browser, *args, **kwargs)

    return inner


class LoginBrowser(PagesBrowser):
    """
    A browser which supports login.
    """

    def __init__(self, username: str, password: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.username = username
        self.password = password

    def do_login(self):
        """
        Abstract method to implement to login on website.

        It is called when a login is needed.
        """
        raise NotImplementedError()

    def do_logout(self):
        """
        Logout from website.

        By default, simply clears the cookies.
        """
        self.session.cookies.clear()


class StatesMixin:
    """
    Mixin to store states of browser.

    It saves and loads a ``state`` dict object. By default it contains the
    current url and cookies, but may be overriden by the subclass to store its
    specific stuff.
    """

    __states__: ClassVar[Tuple[str, ...]] = ()
    """
    Saved state variables.
    """

    STATE_DURATION: ClassVar[int | float | None] = None
    """
    In minutes, used to set an expiration datetime object of the state.
    """

    def locate_browser(self, state: dict):
        """
        From the ``state`` object, go on the saved url.
        """
        try:
            self.location(state['url'])
        except (requests.exceptions.HTTPError, requests.exceptions.TooManyRedirects):
            pass

    def _load_cookies(self, cookie_state: str):
        try:
            uncompressed = zlib.decompress(base64.b64decode(cookie_state))
        except (TypeError, zlib.error, EOFError, ValueError):
            self.logger.error('Unable to uncompress cookies from storage')
            return

        try:
            jcookies = json.loads(uncompressed)
        except ValueError:
            self.logger.error('Unable to reload cookies from storage')
        else:
            for jcookie in jcookies:
                self.session.cookies.set(**jcookie)
            self.logger.debug('Reloaded cookies from storage')

    def load_state(self, state: dict):
        """
        Supply a ``state`` object and load it.
        """
        expire = state.get('expire')

        if expire:
            expire = parser.parse(expire)
            if not expire.tzinfo:
                expire = expire.replace(tzinfo=tz.tzlocal())
            if expire < now_as_utc():
                self.logger.info('State expired, not reloading it from storage')
                return

        if 'cookies' in state:
            self._load_cookies(state['cookies'])

        for attrname in self.__states__:
            if attrname in state:
                setattr(self, attrname, state[attrname])

        if 'url' in state:
            self.locate_browser(state)

    def get_expire(self) -> str | None:
        """
        Get expiration of the ``state`` object, using the :attr:`STATE_DURATION` class attribute.
        """
        if self.STATE_DURATION is None:
            return None

        return str((now_as_utc() + timedelta(minutes=self.STATE_DURATION)).replace(microsecond=0))

    def dump_state(self) -> dict:
        """
        Dump the current state in a ``state`` object.

        Can be overloaded by the browser subclass.
        """
        # XXX similar to Browser.export_session, should be merged, or use it?
        state = {}
        if hasattr(self, 'page') and self.page:
            state['url'] = self.page.url

        cookies = [
            {
                attr: getattr(cookie, attr)
                for attr in ('name', 'value', 'domain', 'path', 'secure', 'expires')
            }
            for cookie in self.session.cookies
        ]
        state['cookies'] = base64.b64encode(zlib.compress(json.dumps(cookies).encode('utf-8'))).decode('ascii')
        for attrname in self.__states__:
            try:
                state[attrname] = getattr(self, attrname)
            except AttributeError:
                pass
        if self.STATE_DURATION is not None:
            state['expire'] = self.get_expire()
        self.logger.debug('Stored cookies into storage')
        return state


class APIBrowser(DomainBrowser):
    """
    A browser for API websites.
    """

    def build_request(self, *args, **kwargs) -> requests.Request:
        if 'data' in kwargs and isinstance(kwargs['data'], dict):
            kwargs['data'] = json.dumps(kwargs['data'])
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['Content-Type'] = 'application/json'

        return super().build_request(*args, **kwargs)

    def open(self, *args, **kwargs) -> requests.Response:
        """
        Do a JSON request.

        The "Content-Type" header is always set to "application/json".

        :param data: if specified, format as JSON and send as request body
        :type data: dict
        :param headers: if specified, add these headers to the request
        :type headers: dict
        """
        return super().open(*args, **kwargs)

    def request(self, *args, **kwargs) -> requests.Response:
        """
        Do a JSON request and parse the response.

        :returns: a dict containing the parsed JSON server response
        :rtype: dict
        """
        return self.open(*args, **kwargs).json()


class AbstractBrowserMissingParentError(Exception):
    """
    .. deprecated:: 3.4
       Don't use this class, import woob_modules.other_module.etc instead
    """


class MetaBrowser(type):
    """
    .. deprecated:: 3.4
       Don't use this class, import woob_modules.other_module.etc instead
    """

    _parent_attr_re = re.compile(r'^[^.]+\.(.*)\.([^.]+)$')

    def __new__(mcs, name, bases, dct):
        from woob.tools.backend import Module  # Here to avoid file wide circular dependency

        if name != 'AbstractBrowser' and AbstractBrowser in bases:
            warnings.warn(
                'AbstractBrowser is deprecated and will be removed in woob 4.0. '
                'Use standard "from woob_modules.other_module import Browser" instead.',
                DeprecationWarning,
                stacklevel=2
            )

            parent_attr = dct.get('PARENT_ATTR')
            if parent_attr:
                m = MetaBrowser._parent_attr_re.match(parent_attr)
                path, klass_name = m.group(1, 2)
                module = importlib.import_module('woob_modules.%s.%s' % (dct['PARENT'], path))
                klass = getattr(module, klass_name)
            else:
                module = importlib.import_module('woob_modules.%s' % dct['PARENT'])
                for attrname in dir(module):
                    attr = getattr(module, attrname)
                    if isinstance(attr, type) and issubclass(attr, Module) and attr != Module:
                        klass = attr.BROWSER
                        break

            bases = tuple(klass if isinstance(base, mcs) else base for base in bases)

        return super().__new__(mcs, name, bases, dct)


class AbstractBrowser(metaclass=MetaBrowser):
    """
    .. deprecated:: 3.4
       Don't use this class, import woob_modules.other_module.etc instead
    """


class OAuth2Mixin(StatesMixin):
    AUTHORIZATION_URI: ClassVar[str, None] = None
    """
    OAuth2 Authorization URI.
    """

    ACCESS_TOKEN_URI: ClassVar[str, None] = None
    """
    OAuth2 route to exchange a code with an access_token.
    """

    SCOPE: ClassVar[str] = ''
    """
    OAuth2 scope.
    """

    client_id: str | None = None
    client_secret: str | None = None
    redirect_uri: str | None = None
    access_token: str | None = None
    access_token_expire: datetime | None = None
    auth_uri: str | None = None
    token_type: str | None = None
    refresh_token: str | None = None
    oauth_state: str | None = None
    authorized_date: str | None = None
    callback_error_description = (
       'operation canceled by the client',
       'login cancelled',
       'consent denied',
       'psu cancelled the transaction',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__states__ += (
            'access_token', 'refresh_token', 'token_type',
        )

    def build_request(self, *args, **kwargs) -> requests.Request:
        headers = kwargs.setdefault('headers', {})
        if self.access_token:
            headers['Authorization'] = '{} {}'.format(self.token_type, self.access_token)
        return super().build_request(*args, **kwargs)

    def dump_state(self) -> dict:
        state = super().dump_state()

        if self.authorized_date:
            state['authorized_date'] = str(self.authorized_date)
        if self.access_token_expire:
            state['access_token_expire'] = self.access_token_expire.isoformat()

        return state

    def load_state(self, state: dict):
        def load_date_or_none(date_str):
            if date_str:
                date_converted = parser.parse(date_str)
                if isinstance(date_converted, datetime):
                    return date_converted
            return None

        super().load_state(state)
        self.authorized_date = load_date_or_none(state.get('authorized_date'))
        self.access_token_expire = load_date_or_none(state.get('access_token_expire'))
        if self.access_token_expire and not self.access_token_expire.tzinfo:
            self.access_token_expire = self.access_token_expire.replace(tzinfo=tz.tzlocal())

    def raise_for_status(self, response: requests.Response):
        if response.status_code == 401:
            self.access_token = None

        return super().raise_for_status(response)

    @property
    def logged(self) -> bool:
        # 'access_token_expire' is already set as UTC in OAuth2Mixin.update_token.
        return self.access_token is not None and (not self.access_token_expire or self.access_token_expire > now_as_utc())

    def do_login(self):
        if self.refresh_token:
            self.use_refresh_token()
        elif self.auth_uri:
            self.request_access_token(self.auth_uri)
        else:
            self.request_authorization()

    def build_authorization_parameters(self) -> dict:
        params = {
            'redirect_uri':    self.redirect_uri,
            'scope':           self.SCOPE,
            'client_id':       self.client_id,
            'response_type':   'code',
        }
        if self.oauth_state:
            params['state'] = self.oauth_state
        return params

    def build_authorization_uri(self) -> str:
        if self.AUTHORIZATION_URI is None:
            self.logger.warning('Use AUTHORIZATION_URI which is None')

        p = urlparse(self.AUTHORIZATION_URI)
        q = dict(parse_qsl(p.query))
        q.update(self.build_authorization_parameters())
        return p._replace(query=urlencode(q)).geturl()

    def request_authorization(self):
        self.logger.info('request authorization')
        raise BrowserRedirect(self.build_authorization_uri())

    def handle_callback_error(self, values: dict):
        callback_error_description = re.compile('|'.join(self.callback_error_description))

        error = values.get('error')
        error_message = values.get('error_description')

        if error == 'access_denied':
            if error_message:
                if callback_error_description.search(error_message.lower()):
                    raise BrowserIncorrectPassword(error_message)
                raise AssertionError(f'Unhandled callback error_message: {error_message}')
            # access_denied with no error message
            raise BrowserIncorrectPassword()
        if error == 'server_error':
            raise BrowserUnavailable()
        raise AssertionError(f'Unhandled callback error: {error}, error_message: {error_message}')

    def build_access_token_parameters(self, values: dict) -> dict:
        return {
            'code': values['code'],
            'grant_type': 'authorization_code',
            'redirect_uri': self.redirect_uri,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }

    def do_token_request(self, data):
        if self.ACCESS_TOKEN_URI is None:
            self.logger.warning('Use ACCESS_TOKEN_URI which is None')

        return self.open(self.ACCESS_TOKEN_URI, data=data)

    def request_access_token(self, auth_uri: str | Dict):
        self.logger.info('requesting access token')

        if isinstance(auth_uri, dict):
            values = auth_uri
        else:
            values = dict(parse_qsl(urlparse(auth_uri).query))
        if not values.get('code'):
            self.handle_callback_error(values)
        self.authorized_date = now_as_utc()
        data = self.build_access_token_parameters(values)
        try:
            auth_response = self.do_token_request(data).json()
        except ClientError:
            raise AssertionError('PSU has successfully logged in, but the request to get an access token failed.')

        self.update_token(auth_response)

    def build_refresh_token_parameters(self) -> dict:
        return {
            'grant_type': 'refresh_token',
            'refresh_token': self.refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri': self.redirect_uri,
        }

    def use_refresh_token(self):
        self.logger.info('refreshing token')

        data = self.build_refresh_token_parameters()
        try:
            auth_response = self.do_token_request(data).json()
        except ClientError:
            self.refresh_token = None
            raise BrowserIncorrectPassword()

        self.update_token(auth_response)

    def update_token(self, auth_response: dict):
        self.token_type = auth_response.get('token_type', 'Bearer').capitalize() # don't know yet if this is a good idea, but required by bnpstet
        if 'refresh_token' in auth_response:
            self.refresh_token = auth_response['refresh_token']
        self.access_token = auth_response['access_token']
        self.access_token_expire = now_as_utc() + timedelta(seconds=int(auth_response['expires_in']))


class OAuth2PKCEMixin(OAuth2Mixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__states__ += ('pkce_verifier', 'pkce_challenge')
        self.pkce_verifier = self.code_verifier()
        self.pkce_challenge = self.code_challenge(self.pkce_verifier)

    # PKCE (Proof Key for Code Exchange) standard protocol methods:
    def code_verifier(self, bytes_number: int = 64) -> str:
        return base64.urlsafe_b64encode(os.urandom(bytes_number)).rstrip(b'=').decode('ascii')

    def code_challenge(self, verifier: str) -> str:
        digest = sha256(verifier.encode('utf8')).digest()
        return base64.urlsafe_b64encode(digest).rstrip(b'=').decode('ascii')

    def build_authorization_parameters(self) -> dict:
        params = {
            'redirect_uri': self.redirect_uri,
            'code_challenge_method': 'S256',
            'code_challenge': self.pkce_challenge,
            'client_id': self.client_id,
        }
        if self.oauth_state:
            params['state'] = self.oauth_state
        return params

    def build_access_token_parameters(self, values: dict) -> dict:
        return {
            'code': values['code'],
            'grant_type': 'authorization_code',
            'code_verifier': self.pkce_verifier,
            'redirect_uri': self.redirect_uri,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }


class DigestMixin:
    """Browser mixin to add a ``Digest`` header compliant with RFC 3230 section 4.3.2."""

    HTTP_DIGEST_ALGORITHM: str = 'SHA-256'
    """Digest algorithm used to obtain a hash of the request content.

    The only supported digest algorithm for now is 'SHA-256'.
    """

    HTTP_DIGEST_METHODS: tuple[str, ...] | None = ('GET', 'POST', 'PUT', 'DELETE')
    """The list of HTTP methods on which to add a ``Digest`` header.

    To add the ``Digest`` header to all methods, set this constant to None.
    """

    HTTP_DIGEST_COMPACT_JSON: bool = False
    """If the content type of the request payload is JSON, compact it first."""

    def compute_digest_header(self, body: bytes) -> str:
        """Compute the value of the ``Digest`` header.

        :param body: The body to compute with.
        :return: The computed ``Digest`` header value.
        """
        if self.HTTP_DIGEST_ALGORITHM == 'SHA-256':
            return 'SHA-256=' + base64.b64encode(sha256(body).digest()).decode()

        raise ValueError(f'Unhandled digest algorithm {self.HTTP_DIGEST_ALGORITHM!r}')

    def add_digest_header(self, preq: requests.PreparedRequest) -> None:
        """Add the ``Digest`` header to the prepared request.

        The ``Digest`` header presence depends on the request:

        - If the request has a ``HTTP_DIGEST_INCLUDE`` header, the ``Digest`` header is added.
        - Otherwise, if the request has a ``HTTP_DIGEST_EXCLUDE`` header, the ``Digest`` header is not added.
        - Otherwise, if :attr:`HTTP_DIGEST_METHOD` is an HTTP method list and the request method is not in said list,
          the ``Digest`` header is not added.
        - Otherwise, the ``Digest`` header is added.

        Note that the ``HTTP_DIGEST_INCLUDE`` and ``HTTP_DIGEST_EXCLUDE`` headers are removed from the request before
        sending it.

        :param preq: The prepared request on which the ``Digest`` header is added.

        .. code-block:: python

            class MyBrowser(DigestMixin, Browser):
                HTTP_DIGEST_METHODS = ('POST', 'PUT', 'DELETE')

            my_browser = MyBrowser()
            my_browser.open('https://example.org/')
        """
        digest_include = preq.headers.pop('HTTP_DIGEST_INCLUDE', None) is not None
        digest_exclude = preq.headers.pop('HTTP_DIGEST_EXCLUDE', None) is not None

        allowed_digest_method = (
            self.HTTP_DIGEST_METHODS is None
            or preq.method in self.HTTP_DIGEST_METHODS
        )

        if not (allowed_digest_method or digest_include) or digest_exclude:
            return

        body = preq.body or b''
        if isinstance(body, str):
            body = body.encode('utf-8')

        if self.HTTP_DIGEST_COMPACT_JSON:
            if 'application/json' in preq.headers.get('Content-Type', ''):
                body = json.dumps(json.loads(body), separators=(',', ':')).encode('utf-8')

        preq.headers['Digest'] = self.compute_digest_header(body)

    def prepare_request(self, *args, **kwargs) -> requests.PreparedRequest:
        """
        Get the prepared request with a ``Digest`` header.
        """
        preq = super().prepare_request(*args, **kwargs)
        self.add_digest_header(preq)
        return preq



# Copyright(C) 2010-2016 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import enum
import re
from datetime import date, datetime
from typing import Iterable, Iterator

from unidecode import unidecode

from woob.capabilities.base import (
    BaseObject, Field, StringField, DecimalField,
    UserError, Currency, EnumField, Enum,
    Capability, empty, find_object,
)
from woob.capabilities.date import DateField
from woob.exceptions import BrowserQuestion
from woob.tools.capabilities.bank.iban import is_iban_valid

from .base import ObjectNotFound, BaseAccount, CapBank, Account


__all__ = [
    'AddRecipientBankError', 'AddRecipientError', 'AddRecipientStep', 'AddRecipientTimeout',
    'BeneficiaryType',
    'CapBankTransfer', 'CapBankTransferAddRecipient',
    'CapTransfer', 'Emitter', 'EmitterNumberType', 'Recipient',
    'RecipientInvalidIban', 'RecipientInvalidLabel', 'RecipientInvalidOTP', 'RecipientNotFound',
    'Transfer',
    'TransferBankError', 'TransferCancelledByUser', 'TransferDateType', 'TransferError', 'TransferFrequency',
    'TransferInsufficientFunds', 'TransferInvalidAmount', 'TransferInvalidCurrency',
    'TransferInvalidDate', 'TransferInvalidEmitter', 'TransferInvalidLabel', 'TransferTimeout', 'TransferInvalidOTP',
    'TransferInvalidRecipient', 'TransferNotFound', 'TransferStatus', 'TransferStep',
]


class RecipientNotFound(ObjectNotFound):
    """
    Raised when a recipient is not found.
    """

    def __init__(self, msg: str = 'Recipient not found'):
        super().__init__(msg)


class TransferNotFound(ObjectNotFound):
    def __init__(self, msg: str = 'Transfer not found'):
        super().__init__(msg)


class TransferError(UserError):
    """
    A transfer has failed.

    :param description: technical description of the error
    :type description: str
    :param message: error message from the bank, if any, to display to the user
    :type message: str
    """

    code = 'transferError'

    def __init__(self, description: str | None = None, message: str | None = None):
        super().__init__(message or description)
        self.message = message
        self.description = description


class TransferBankError(TransferError):
    """The transfer was rejected by the bank with a message."""

    code = 'bankMessage'


class TransferTimeout(TransferError):
    """The transfer request timed out"""

    code = 'timeout'


class TransferInvalidLabel(TransferError):
    """The transfer label is invalid."""

    code = 'invalidLabel'


class TransferInvalidEmitter(TransferError):
    """The emitter account cannot be used for transfers."""

    code = 'invalidEmitter'


class TransferInvalidRecipient(TransferError):
    """The emitter cannot transfer to this recipient."""

    code = 'invalidRecipient'


class TransferInvalidAmount(TransferError):
    """This amount is not allowed."""

    code = 'invalidAmount'


class TransferInvalidCurrency(TransferInvalidAmount):
    """The transfer currency is invalid."""

    code = 'invalidCurrency'


class TransferInsufficientFunds(TransferInvalidAmount):
    """Not enough funds on emitter account."""

    code = 'insufficientFunds'


class TransferInvalidDate(TransferError):
    """This execution date cannot be used."""

    code = 'invalidDate'


class TransferInvalidOTP(TransferError):
    code = 'invalidOTP'


class TransferCancelledByUser(TransferError):
    """The transfer is cancelled by the emitter or an authorized user"""

    code = 'cancelledByUser'


class AddRecipientError(UserError):
    """
    Failed trying to add a recipient.

    :param description: technical description of the error
    :type description: str
    :param message: error message from the bank, if any, to display to the user
    :type message: str
    """

    code = 'AddRecipientError'

    def __init__(self, description: str | None = None, message: str | None = None):
        super().__init__(message or description)
        self.message = message
        self.description = description


class AddRecipientBankError(AddRecipientError):
    """The new recipient was rejected by the bank with a message."""

    code = 'bankMessage'


class AddRecipientTimeout(AddRecipientError):
    """Add new recipient request has timeout"""

    code = 'timeout'


class RecipientInvalidIban(AddRecipientError):
    code = 'invalidIban'


class RecipientInvalidLabel(AddRecipientError):
    code = 'invalidLabel'


class RecipientInvalidOTP(AddRecipientError):
    code = 'invalidOTP'


class Recipient(BaseAccount):
    """
    Recipient of a transfer.
    """
    enabled_at =     DateField('Date of availability')
    category =       StringField('Recipient category')
    iban =           StringField('International Bank Account Number')

    # Needed for multispaces case
    origin_account_id = StringField('Account id which recipient belong to')
    origin_account_iban = StringField('Account iban which recipient belong to')

    def __repr__(self):
        return "<%s id=%r label=%r>" % (type(self).__name__, self.id, self.label)


class TransferStep(BrowserQuestion):
    # TODO doc
    def __init__(self, transfer: Transfer, *values):
        super().__init__(*values)
        self.transfer = transfer


class AddRecipientStep(BrowserQuestion):
    # TODO doc
    def __init__(self, recipient: Recipient, *values):
        super().__init__(*values)
        self.recipient = recipient


class BeneficiaryType(Enum):
    """
    Type of the Transfer.beneficiary_number property.
    """

    IBAN = 'iban'
    """beneficiary number is an IBAN as defined in ISO 13616"""

    SORT_CODE_ACCOUNT_NUMBER = 'sort_code_account_number'
    """account number is a national UK/Ireland number including sortcode"""

    PHONE_NUMBER = 'phone_number'
    """beneficiary number is an E.164 encoded phone number"""

    RECIPIENT = 'recipient'
    """
    beneficiary number is a beneficiary identifier as returned by
    the iter_transfer_recipients method.
    """


class TransferStatus(Enum):
    UNKNOWN = 'unknown'

    INTERMEDIATE = 'intermediate'
    """Transfer is not validated yet"""

    SCHEDULED = 'scheduled'
    """Transfer to be executed later"""

    ACTIVE = 'active'
    """Periodic transfer is still active"""

    DONE = 'done'
    """Transfer was executed"""

    CANCELLED = 'cancelled'
    """Transfer was cancelled by the bank or by the user"""

    ACCEPTED_NO_BANK_STATUS = 'accepted_no_bank_status'
    """Transfer was sent to the bank but we will not get more information
    after that. This is used for banks that do not give us final states after pending."""


class TransferFrequency(Enum):
    UNKNOWN = 'unknown'
    DAILY = 'daily'
    WEEKLY = 'weekly'
    TWOWEEKLY = 'two-weekly'  # every two weeks, not 2 times per week
    MONTHLY = 'monthly'
    TWOMONTHLY  = 'two-monthly'  # every two months, not 2 times per month
    QUARTERLY = 'quarterly'
    FOURMONTHLY = 'four-monthly'  # every four months, not 4 times per month
    SEMIANNUALLY = 'semiannually'
    YEARLY = 'yearly'

    # (deprecated)
    BIMONTHLY = 'bimonthly'  # use TWOWEEKLY instead
    BIANNUAL = 'biannual'  # use SEMIANNUALLY instead


class TransferDateType(Enum):
    FIRST_OPEN_DAY = 'first_open_day'
    """Transfer to execute when possible (accounting opening days)"""

    INSTANT = 'instant'
    """Transfer to execute immediately (not accounting opening days)"""

    DEFERRED = 'deferred'
    """Transfer to execute on a chosen date"""

    PERIODIC = 'periodic'
    """Transfer to execute periodically"""


class Transfer(BaseObject, Currency):
    """
    Transfer from an account to a recipient.
    """
    amount =          DecimalField('Amount to transfer')
    currency =        StringField('Currency', default=None)
    fees =            DecimalField('Fees', default=None)

    exec_date =       Field('Date of transfer', date, datetime)
    label =           StringField('Reason')

    account_id =      StringField('ID of origin account')
    account_iban =    StringField('International Bank Account Number')
    account_label =   StringField('Label of origin account')
    account_balance = DecimalField('Balance of origin account before transfer')

    # Information for beneficiary in recipient list
    recipient_id =      StringField('ID of recipient account')
    recipient_iban =    StringField('International Bank Account Number')
    recipient_label =   StringField('Label of recipient account')

    # Information for beneficiary not only in recipient list
    # Like transfer to iban beneficiary
    beneficiary_type =    StringField('Transfer creditor number type', default=BeneficiaryType.RECIPIENT)
    beneficiary_number =  StringField('Transfer creditor number')
    beneficiary_label =  StringField('Transfer creditor label')
    beneficiary_bic = StringField('Transfer creditor BIC')

    date_type = EnumField('Transfer execution date type', TransferDateType)

    frequency = EnumField('Frequency of periodic transfer', TransferFrequency)
    first_due_date = DateField('Date of first transfer of periodic transfer')
    last_due_date = DateField('Date of last transfer of periodic transfer')

    creation_date = DateField('Creation date of transfer')
    status = EnumField('Transfer status', TransferStatus)

    cancelled_exception = Field('Transfer cancelled reason', TransferError)

    # End to end ID given by the client
    reference_id = StringField('End to end ID given by client')


class EmitterNumberType(Enum):
    UNKNOWN = 'unknown'
    IBAN = 'iban'
    BBAN = 'bban'


class Emitter(BaseAccount):
    """
    Transfer emitter account.
    """
    number_type = EnumField('Account number type', EmitterNumberType, default=EmitterNumberType.UNKNOWN)
    number = StringField('Account number value')
    balance = DecimalField('Balance of emitter account')


class DebtorAccountRequirement(Enum):
    MANDATORY = 'mandatory'
    """Debtor account is needed to initiate a payment"""

    OPTIONAL = 'optional'
    """Debtor account is optional (may change module behaviour)"""

    NOT_USED = 'not_used'
    """Debtor account must not be given"""


class Platform(str, enum.Enum):
    """Mobile platform on which the webview can be run.

    For instance, this enumeration can be used to represent systems on which
    the authorization link can be catched by a native application instead of
    the browser.
    """

    ANDROID = 'android'
    """Android based platforms."""

    IOS = 'ios'
    """Apple's iOS platform."""


class CapTransfer(Capability):
    can_do_transfer_to_untrusted_beneficiary = False
    """
    The module can do transfer to untrusted beneficiary, for example:
    when module can't add new beneficiary without doing a transfer like n26
    or when module can do transfer to a beneficiary not listed
    in `iter_transfer_recipients` like for PSD2 modules
    """

    can_do_transfer_without_emitter = True
    """
    The module can do transfer without giving the emitter, for example:
    when there is only, and will be only, one account like wallet
    or when the module can initiate transfer without emitter
    and the emitter is chosen afterwards like for PSD2 modules
    """

    can_do_transfer_cancellation = False

    sca_required_for_transfer_cancellation = False
    """
    The default behavior is that we don't need to validate a payment cancellation through a SCA.
    If a SCA is required after sending a transfer cancellation request to validate it, the module
    should set this to True.
    """

    accepted_beneficiary_types = (BeneficiaryType.RECIPIENT, )
    accepted_execution_date_types = (TransferDateType.FIRST_OPEN_DAY, TransferDateType.DEFERRED)
    accepted_execution_frequencies = set(TransferFrequency) - set([TransferFrequency.UNKNOWN])
    maximum_number_of_instructions = 1
    transfer_with_debtor_account = DebtorAccountRequirement.NOT_USED

    # Indicate that we may not know if the payment is done or rejected, the information
    # is provided by date type because this behaviour is generally dependent on the type
    # of payment. An empty list means that the transfer will never get the status ACCEPTED_NO_BANK_STATUS
    partial_transfer_status_tracking = ()

    is_app_to_app_used_for_transfer = {
        Platform.ANDROID: None,
        Platform.IOS: None,
    }  # type: dict[Platform, bool | None]
    """
    Is an App2App flow used for the payment if the PSU has the bank's app installed.
    None means unknown
    """

    bank_provides_payer_account = None  # type: bool | None
    """
    Once the payment is initiated, does the bank return the payer's account identifier?
    None means unknown
    """

    bank_provides_payer_label = None  # type: bool | None
    """
    Once the payment is initiated, does the bank return the payer's label?
    None means unknown
    """

    transfer_date_types_where_trusted_beneficiary_required = set()  # type: Iterable[TransferDateType]
    """
    Set of `TransferDateType` where the beneficiary must be trusted or registered on the payer's banking service.
    If `iter_transfer_recipients` is implemented, such beneficiaries may be found.
    """

    def iter_transfer_recipients(self, account: Account) -> Iterator[Recipient]:
        """
        Iter recipients availables for a transfer from a specific account.

        :param account: account which initiate the transfer
        :type account: :class:`Account`
        :rtype: iter[:class:`Recipient`]
        :raises: :class:`AccountNotFound`
        """
        raise NotImplementedError()

    def init_transfer(self, transfer: Transfer, **params) -> Transfer:
        """
        Initiate a transfer.

        :param :class:`Transfer`
        :rtype: :class:`Transfer`
        :raises: :class:`TransferError`
        """
        raise NotImplementedError()

    def execute_transfer(self, transfer: Transfer, **params) -> Transfer:
        """
        Execute a transfer.

        :param :class:`Transfer`
        :rtype: :class:`Transfer`
        :raises: :class:`TransferError`
        """
        raise NotImplementedError()

    def confirm_transfer(self, transfer: Transfer, **params) -> Transfer:
        """
        Transfer confirmation after multiple SCA from the Emitter.
        This method is only used for PSD2 purpose.
        Return the transfer with the new status.

        :param :class:`Transfer`
        :rtype: :class:`Transfer`
        :raises: :class:`TransferError`
        """
        return self.get_transfer(transfer.id)

    def confirm_transfer_cancellation(self, transfer: Transfer, **params) -> Transfer:
        """
        Confirm transfer cancellation after a redirect flow.

        :param :class:`Transfer`
        :rtype: :class:`Transfer`
        :raises: :class:`AssertionError`: If the payment is not actually cancelled after the whole process
        """
        transfer = self.optional_confirm_transfer_cancellation(transfer, **params)
        # Check that the transfer has been successfully cancelled.
        if transfer.status != TransferStatus.CANCELLED:
            raise AssertionError('Transfer is not cancelled after cancellation request.')
        return transfer

    def optional_confirm_transfer_cancellation(self, transfer: Transfer, **params) -> Transfer:
        """Proceed with the actual cancellation confirmation.

        This method MUST NOT be called by any external caller. Said caller
        should actually call confirm_transfer_cancellation which may call the
        current method if it sees fit.

        The default implementation does not run an explicit confirmation step,
        it only fetches the up-to-date transfer.

        Modules requiring an explicit cancellation confirmation should
        overwrite this method, returning the up-to-date transfer at the end.
        """
        return self.get_transfer(transfer.id)

    def transfer(self, transfer: Transfer, **params) -> Transfer:
        """
        Do a transfer from an account to a recipient.

        :param :class:`Transfer`
        :rtype: :class:`Transfer`
        :raises: :class:`TransferError`
        """

        transfer_not_check_fields = {
            BeneficiaryType.RECIPIENT: ('id', 'beneficiary_number', 'beneficiary_label',),
            BeneficiaryType.IBAN: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
            BeneficiaryType.PHONE_NUMBER: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
            BeneficiaryType.SORT_CODE_ACCOUNT_NUMBER: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
        }

        if hasattr(transfer, 'instructions'):
            instructions = transfer.instructions
        else:
            instructions = [transfer]
        nb_instructions = len(instructions)

        for instr in instructions:
            if not instr.amount or instr.amount <= 0:
                raise TransferInvalidAmount('amount must be strictly positive')

        instructions = sorted(
            instructions,
            key=lambda x: (x.reference_id, x.beneficiary_number, x.recipient_iban, x.amount, x.exec_date)
        )

        # Initiate the transfer
        t = self.init_transfer(transfer, **params)

        # Verify the created transfer before execution
        if hasattr(t, 'instructions'):
            new_instructions = t.instructions
        else:
            new_instructions = [t]
        nb_new_instructions = len(new_instructions)
        new_instructions = sorted(
            new_instructions,
            key=lambda x: (x.reference_id, x.beneficiary_number, x.recipient_iban, x.amount, x.exec_date)
        )

        assert nb_instructions == nb_new_instructions, (
            'Number of instructions changed during transfer processing (from "%s" to "%s")' % (nb_instructions, nb_new_instructions)
        )

        changed_msg_template = '%s changed during transfer processing (from "%s" to "%s") for instruction %s'
        for orig_instr, new_instr in zip(instructions, new_instructions):
            ignored_keys = transfer_not_check_fields[orig_instr.beneficiary_type]
            for key, value in new_instr.iter_fields():
                if key in ignored_keys:
                    continue
                try:
                    transfer_val = getattr(orig_instr, key)
                except AttributeError:
                    continue
                transfer_check_fn = getattr(self, 'transfer_check_%s' % key, None)
                if transfer_check_fn:
                    if not transfer_check_fn(transfer_val, value):
                        raise AssertionError(changed_msg_template % (
                            key, transfer_val, value, orig_instr.reference_id or ''
                        ))
                elif transfer_val != value and not empty(transfer_val):
                    raise AssertionError(changed_msg_template % (
                        key, transfer_val, value, orig_instr.reference_id or ''
                    ))
        return self.execute_transfer(t, **params)

    def transfer_check_label(self, old: str, new: str) -> bool:
        old = re.sub(r'\s+', ' ', old).strip()
        new = re.sub(r'\s+', ' ', new).strip()
        return unidecode(old) == unidecode(new)

    def iter_transfers(self, account: Account | None = None) -> Iterator[Transfer]:
        """
        Iter transfer transactions.

        :param account: account to get transfer history (or None for all accounts)
        :type account: :class:`Account`
        :rtype: iter[:class:`Transfer`]
        :raises: :class:`AccountNotFound`
        """
        raise NotImplementedError()

    def get_transfer(self, id: str) -> Transfer:
        """
        Get a transfer from its id.

        :param id: ID of the Transfer
        :type id: :class:`str`
        :rtype: :class:`Transfer`
        """
        return find_object(self.iter_transfers(), id=id, error=TransferNotFound)

    def iter_emitters(self) -> Iterator[Emitter]:
        """
        Iter transfer emitter accounts.

        :rtype: iter[:class:`Emitter`]
        """
        raise NotImplementedError()

    def cancel_transfer(self, transfer: Transfer, **params) -> Transfer:
        """
        Ask for the cancellation of a transfer.

        This function is exposed as part of Woob API and should not be overriden
        by children modules.

        :param transfer: the transfer that should be cancelled
        :type transfer: :class:`Transfer`
        :rtype: :class:`Transfer`
        """
        return self.do_transfer_cancellation(transfer, **params)

    def do_transfer_cancellation(self, transfer: Transfer, **params) -> Transfer:
        """
        Send a cancellation request for the given transfer.

        This function should be implemented by the children modules.

        :param transfer: the transfer that should be cancelled
        :type transfer: :class:`Transfer`
        :rtype: :class:`Transfer`
        """
        raise NotImplementedError()


class CapBankTransfer(CapBank, CapTransfer):
    can_do_transfer_without_emitter = False
    transfer_with_debtor_account = DebtorAccountRequirement.MANDATORY

    def account_to_emitter(self, account: str | Account) -> Emitter:
        if isinstance(account, Account):
            id = account.id
        else:
            id = account

        return find_object(self.iter_emitters(), id=id, error=ObjectNotFound)


class CapBankTransferAddRecipient(CapBankTransfer):
    def new_recipient(self, recipient: Recipient, **params):
        # TODO doc
        raise NotImplementedError()

    def add_recipient(self, recipient: Recipient, **params) -> Recipient:
        """
        Add a recipient to the connection.

        :param recipient: recipient to add, need to contain a valid IBAN and a label.
        :raises: :class:`BrowserQuestion`
        :raises: :class:`AddRecipientError`
        :rtype: :class:`Recipient`
        """
        if not is_iban_valid(recipient.iban):
            raise RecipientInvalidIban('Iban is not valid.')
        if not recipient.label:
            raise RecipientInvalidLabel('Recipient label is mandatory.')
        return self.new_recipient(recipient, **params)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.geolocip import CapGeolocIp
from woob.tools.application.repl import ReplApplication


__all__ = ['AppGeolocIP']


class AppGeolocIP(ReplApplication):
    APPNAME = 'geolocip'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to geolocalize IP addresses."
    SHORT_DESCRIPTION = "geolocalize IP addresses"
    CAPS = CapGeolocIp

    def main(self, argv):
        if len(argv) < 2:
            print('Syntax: %s ipaddr' % argv[0], file=self.stderr)
            return 2

        for location in self.do('get_location', argv[1]):
            if location.lt and location.lg:
                location.osmlink = 'https://www.openstreetmap.org/?mlat=%s&mlon=%s#map=13/%s/%s' % (location.lt, location.lg, location.lt, location.lg)
            self.format(location)



# Copyright(C) 2012-2016 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

__all__ = ['LimitedLRUDict', 'LRUDict']


class LRUDict(OrderedDict):
    """dict to store items in the order the keys were last added/fetched."""

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        super(LRUDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        value = super(LRUDict, self).__getitem__(key)
        self[key] = value
        return value


class LimitedLRUDict(LRUDict):
    """dict to store only the N most recent items."""

    max_entries = 100

    def __setitem__(self, key, value):
        super(LimitedLRUDict, self).__setitem__(key, value)
        if len(self) > self.max_entries:
            self.popitem(last=False)



# Copyright(C) 2012 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, Field, DecimalField, \
                  StringField, UserError
from .date import DateField

__all__ = ['Shop', 'Price', 'Product', 'CapPriceComparison']


class PriceNotFound(UserError):
    """
    Raised when a price is not found
    """

    def __init__(self, msg='Price not found'):
        super(PriceNotFound, self).__init__(msg)


class Product(BaseObject):
    """
    A product.
    """
    name =      StringField('Name of product')


class Shop(BaseObject):
    """
    A shop where the price is.
    """
    name =      StringField('Name of shop')
    location =  StringField('Location of the shop')
    info =      StringField('Information about the shop')


class Price(BaseObject):
    """
    Price.
    """
    date =      DateField('Date when this price has been published')
    cost =      DecimalField('Cost of the product in this shop')
    currency =  StringField('Currency of the price')
    message =   StringField('Message related to this price')
    shop =      Field('Shop information', Shop)
    product =   Field('Product', Product)


class CapPriceComparison(Capability):
    """
    Capability for price comparison websites.
    """

    def search_products(self, pattern=None):
        """
        Search products from a pattern.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`Product`]
        """
        raise NotImplementedError()

    def iter_prices(self, products):
        """
        Iter prices for a product.

        :param product: product to search
        :type product: :class:`Product`
        :rtype: iter[:class:`Price`]
        """
        raise NotImplementedError()

    def get_price(self, id):
        """
        Get a price from an ID

        :param id: ID of price
        :type id: str
        :rtype: :class:`Price`
        """
        raise NotImplementedError()



# Copyright(C) 2017  Vincent A
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
from functools import wraps

from .browsers import LoginBrowser
from .exceptions import LoggedOut
from ..exceptions import BrowserUnavailable


__all__ = ['login_method', 'retry_on_logout', 'RetryLoginBrowser']


def login_method(func):
    """Decorate a method to indicate the browser is logging in.

    When the decorated method is called, pages like
    `woob.browser.pages.LoginPage` will not raise `LoggedOut`, since it is
    expected for the browser not to be logged yet.
    """

    func.login_decorated = True

    @wraps(func)
    def wrapper(browser, *args, **kwargs):
        browser.logging_in += 1
        try:
            return func(browser, *args, **kwargs)
        finally:
            browser.logging_in -= 1
    return wrapper


def retry_on_logout(exc_check=LoggedOut, tries=4):
    """Decorate a function to retry several times in case of exception.

    The decorated function is called at max 4 times. It is retried only when it
    raises an exception of the type `woob.browser.exceptions.LoggedOut`.
    If the function call succeeds and returns an iterator, a wrapper to the
    iterator is returned. If iterating on the result raises a `LoggedOut`,
    the iterator is recreated by re-calling the function, but the values
    already yielded will not be re-yielded.
    For consistency, the function MUST always return values in the same order.

    Adding this decorator to a method which can be called from another
    decorated method should be avoided, since nested calls will greatly
    increase the number of retries.
    """

    if not isinstance(exc_check, type) or not issubclass(exc_check, Exception):
        raise TypeError('retry_on_logout() must be called in order to decorate %r' % tries)

    def decorator(func):
        @wraps(func)
        def wrapper(browser, *args, **kwargs):
            cb = lambda: func(browser, *args, **kwargs)

            for i in range(tries, 0, -1):
                try:
                    ret = cb()
                except exc_check as exc:
                    browser.logger.info('%r raised, retrying', exc)
                    continue

                if not (hasattr(ret, '__next__') or hasattr(ret, 'next')):
                    return ret  # simple value, no need to retry on items
                return iter_retry(cb, value=ret, remaining=i, exc_check=exc_check, logger=browser.logger)

            raise BrowserUnavailable('Site did not reply successfully after multiple tries')

        return wrapper
    return decorator


@contextmanager
def retry_on_logout_context(tries=4, logger=None):
    for i in range(tries, 0, -1):
        try:
            yield
        except LoggedOut as exc:
            if logger:
                logger.debug('%r raised, retrying', exc)
        else:
            return
    raise BrowserUnavailable('Site did not reply successfully after multiple tries')


class RetryLoginBrowser(LoginBrowser):
    """Browser that can retry methods if the site logs out the session.

    Some sites can terminate a session anytime, redirecting to a login page.
    To avoid having to handle it in the middle of every method, one can simply
    let logouts raise a `woob.browser.exceptions.LoggedOut` exception that
    is handled with a retry, thanks to the `@retry_on_logout` decorator.

    The `woob.browser.pages.LoginPage` will raise `LoggedOut` if the browser
    is not currently logging in. To detect this situation, the `do_login`
    method MUST be decorated with `@login_method`.
    """
    def __init__(self, *args, **kwargs):
        super(RetryLoginBrowser, self).__init__(*args, **kwargs)
        self.logging_in = 0

        if not hasattr(self.do_login, 'login_decorated'):
            raise Exception('do_login method was not decorated with @login_method')


class iter_retry:
    # when the callback is retried, it will create a new iterator, but we may already yielded
    # some values, so we need to keep track of them and seek in the middle of the iterator

    def __init__(self, cb, remaining=4, value=None, exc_check=LoggedOut, logger=None):
        self.cb = cb
        self.it = iter(value) if value is not None else None
        self.items = []
        self.remaining = remaining
        self.logger = logger
        self.exc_check = exc_check

    def __iter__(self):
        return self

    def __next__(self):
        if self.remaining <= 0:
            raise BrowserUnavailable('Site did not reply successfully after multiple tries')

        if self.it is None:
            self.it = iter(self.cb())

            # recreated iterator, consume previous items
            try:
                nb = -1
                for nb, sent in enumerate(self.items):
                    new = next(self.it)
                    if hasattr(new, 'iter_fields'):
                        equal = dict(sent.iter_fields()) == dict(new.iter_fields())
                    else:
                        equal = sent == new
                    if not equal:
                        # safety is not guaranteed
                        raise BrowserUnavailable('Site replied inconsistently between retries, %r vs %r', sent, new)
            except StopIteration:
                raise BrowserUnavailable('Site replied fewer elements (%d) than last iteration (%d)', nb + 1, len(self.items))
            except self.exc_check as exc:
                if self.logger:
                    self.logger.info('%r raised, retrying', exc)
                self.it = None
                self.remaining -= 1
                return next(self)

        # return one item
        try:
            obj = next(self.it)
        except self.exc_check as exc:
            if self.logger:
                self.logger.info('%r raised, retrying', exc)
            self.it = None
            self.remaining -= 1
            return next(self)
        else:
            self.items.append(obj)
            return obj

    next = __next__



# Copyright(C) 2012-2013 Florent Fourcot
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal

from woob.capabilities.bill import CapDocument, Detail, Subscription
from woob.capabilities.profile import CapProfile
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import PrettyFormatter
from woob.tools.application.base import MoreResultsAvailable
from woob.tools.application.captcha import CaptchaMixin
from woob.core import CallErrors
from woob.exceptions import CaptchaQuestion
from woob.capabilities.captcha import exception_to_job


__all__ = ['AppBill']


class SubscriptionsFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'label')

    def get_title(self, obj):
        if obj.renewdate:
            return "%s - %s" % (obj.label, obj.renewdate.strftime('%d/%m/%y'))
        return obj.label


class AppBill(CaptchaMixin, ReplApplication):
    APPNAME = 'bill'
    OLD_APPNAME = 'boobill'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Florent Fourcot'
    DESCRIPTION = 'Console application allowing to get/download documents and bills.'
    SHORT_DESCRIPTION = "get/download documents and bills"
    CAPS = CapDocument
    COLLECTION_OBJECTS = (Subscription, )
    EXTRA_FORMATTERS = {'subscriptions':   SubscriptionsFormatter,
                        }
    DEFAULT_FORMATTER = 'table'
    COMMANDS_FORMATTERS = {'subscriptions':   'subscriptions',
                           'ls':              'subscriptions',
                          }

    def load_default_backends(self):
        self.load_backends(CapDocument, storage=self.create_storage())

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def exec_method(self, id, method):
        l = []
        id, backend_name = self.parse_id(id)

        if not id:
            for subscrib in self.get_object_list('iter_subscription'):
                l.append((subscrib.id, subscrib.backend))
        else:
            l.append((id, backend_name))

        more_results = []
        not_implemented = []
        self.start_format()
        for id, backend in l:
            names = (backend,) if backend is not None else None
            try:
                for result in self.do(method, id, backends=names):
                    self.format(result)
            except CallErrors as errors:
                for backend, error, backtrace in errors:
                    if isinstance(error, MoreResultsAvailable):
                        more_results.append(id + '@' + backend.name)
                    elif isinstance(error, NotImplementedError):
                        if backend not in not_implemented:
                            not_implemented.append(backend)
                    else:
                        self.bcall_error_handler(backend, error, backtrace)

        if len(more_results) > 0:
            print('Hint: There are more results available for %s (use option -n or count command)' % (', '.join(more_results)), file=self.stderr)
        for backend in not_implemented:
            print('Error(%s): This feature is not supported yet by this backend.' % backend.name, file=self.stderr)

    def do_subscriptions(self, line):
        """
        subscriptions

        List all subscriptions.
        """
        return self.do_ls(line)

    def do_details(self, id):
        """
        details [ID]

        Get details of subscriptions.
        If no ID given, display all details of all backends.
        """
        l = []
        id, backend_name = self.parse_id(id)

        if not id:
            for subscrib in self.get_object_list('iter_subscription'):
                l.append((subscrib.id, subscrib.backend))
        else:
            l.append((id, backend_name))

        for id, backend in l:
            names = (backend,) if backend is not None else None
            # XXX: should be generated by backend? -Flo
            # XXX: no, but you should do it in a specific formatter -romain
            # TODO: do it, and use exec_method here. Code is obsolete
            mysum = Detail()
            mysum.label = "Sum"
            mysum.infos = "Generated by bill"
            mysum.price = Decimal("0.")

            self.start_format()
            for detail in self.do('get_details', id, backends=names):
                self.format(detail)
                mysum.price = detail.price + mysum.price

            self.format(mysum)

    def do_balance(self, id):
        """
        balance [ID]

        Get balance of subscriptions.
        If no ID given, display balance of all backends.
        """

        self.exec_method(id, 'get_balance')

    @defaultcount(10)
    def do_history(self, id):
        """
        history [ID]

        Get the history of subscriptions.
        If no ID given, display histories of all backends.
        """
        self.exec_method(id, 'iter_bills_history')

    @defaultcount(10)
    def do_documents(self, id):
        """
        documents [ID]

        Get the list of documents for subscriptions.
        If no ID given, display documents of all backends
        """
        self.exec_method(id, 'iter_documents')

    @defaultcount(10)
    def do_bills(self, id):
        """
        bills [ID]

        Get the list of bills documents for subscriptions.
        If no ID given, display bills of all backends
        """
        self.exec_method(id, 'iter_bills')

    def do_download(self, line, force_pdf=False):
        """
        download [DOC_ID | all] [FILENAME]

        download DOC_ID [FILENAME]

        download the document
        DOC_ID is the identifier of the document (hint: try documents command)
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.

        download all [SUB_ID]

        You can use special word "all" and download all documents of
        subscription identified by SUB_ID.
        If SUB_ID is not given, download documents of all subscriptions.
        """
        id, dest = self.parse_command_args(line, 2, 1)
        id, backend_name = self.parse_id(id)
        if not id:
            print('Error: please give a document ID (hint: use documents command)', file=self.stderr)
            return 2

        if id == 'all':
            return self.download_all(dest, force_pdf)

        names = (backend_name,) if backend_name is not None else None

        document, = self.do('get_document', id, backends=names)
        if not document:
            print('Error: document not found')
            return 1

        if dest is None:
            extension = document.format if not force_pdf else 'pdf'
            dest = document.id + (f'.{extension}' if extension else '')

        for buf in self.do('download_document' if not force_pdf else 'download_document_pdf', document, backends=names):
            if buf:
                if dest == "-":
                    self.stdout.buffer.write(buf)
                else:
                    try:
                        with open(dest, 'wb') as f:
                            f.write(buf)
                        if not document.has_file:
                            print('Warning: document.has_file is falsy but the file is available', file=self.stderr)
                    except IOError as e:
                        print('Unable to write document in "%s": %s' % (dest, e), file=self.stderr)
                        return 1
                return

    def do_download_pdf(self, line):
        """
        download_pdf [id | all]

        download function with forced PDF conversion.
        """

        return self.do_download(line, force_pdf=True)

    def download_all(self, sub_id, force_pdf):
        if sub_id:
            sub_id, backend_name = self.parse_id(sub_id)
            names = (backend_name,) if backend_name else None
            subscription, = self.do('get_subscription', sub_id, backends=names)
            if not self.download_subscription(subscription, force_pdf):
                return 1
        else:
            for subscription in self.do('iter_subscription'):
                if not self.download_subscription(subscription, force_pdf):
                    return 1

    def download_subscription(self, subscription, force_pdf):
        for document in self.do('iter_documents', subscription, backends=(subscription.backend,)):
            if not self.download_doc(document, force_pdf):
                return False
        return True

    def download_doc(self, document, force_pdf):
        if force_pdf:
            method = 'download_document_pdf'
        else:
            method = 'download_document'

        extension = document.format if not force_pdf else 'pdf'
        dest = document.id + (f'.{extension}' if extension else '')

        for buf in self.do(method, document, backends=(document.backend,)):
            if buf:
                try:
                    with open(dest, 'wb') as f:
                        f.write(buf)
                        if not document.has_file:
                            print('Warning: document.has_file is falsy but the file is available', file=self.stderr)
                except IOError as e:
                    print('Unable to write bill in "%s": %s' % (dest, e), file=self.stderr)
                    return False
        return True

    def do_profile(self, line):
        """
        profile

        Display detailed information about person or company.
        """
        self.start_format()
        for profile in self.do('get_profile', caps=CapProfile):
            self.format(profile)

    def bcall_error_handler(self, backend, error, backtrace):
        """
        Handler for an exception inside the CallErrors exception.

        This method can be overridden to support more exceptions types.
        """
        if isinstance(error, CaptchaQuestion):
            if not self.captcha_woob.count_backends():
                print('Error(%s): Site requires solving a CAPTCHA but no CapCaptchaSolver backends were configured' % backend.name,
                      file=self.stderr)
                return False

            print('Info(%s): Encountered CAPTCHA, please wait for its resolution, it can take dozens of seconds' % backend.name, file=self.stderr)
            job = exception_to_job(error)
            self.solve_captcha(job, backend)
            return False

        return super(ReplApplication, self).bcall_error_handler(backend, error, backtrace)



# Copyright(C) 2012-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .browsers import (
    Browser, DomainBrowser, UrlNotAllowed, PagesBrowser,
    LoginBrowser, need_login, AbstractBrowser, StatesMixin,
    APIBrowser, OAuth2Mixin, OAuth2PKCEMixin,
)
from .url import URL


__all__ = [
    'Browser', 'DomainBrowser', 'UrlNotAllowed', 'PagesBrowser', 'URL',
    'LoginBrowser', 'need_login', 'AbstractBrowser', 'StatesMixin',
    'APIBrowser', 'OAuth2Mixin', 'OAuth2PKCEMixin',
]



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import sys
from collections import defaultdict
from logging import addLevelName, Formatter, getLogger as _getLogger, LoggerAdapter


__all__ = ['getLogger', 'createColoredFormatter', 'settings']


RESET_SEQ = "\033[0m"
COLOR_SEQ = "%s%%s" + RESET_SEQ

COLORS = {
    'DEBUG': COLOR_SEQ % "\033[0;36m",
    'INFO': COLOR_SEQ % "\033[32m",
    'WARNING': COLOR_SEQ % "\033[1;33m",
    'ERROR': COLOR_SEQ % "\033[1;31m",
    'CRITICAL': COLOR_SEQ % ("\033[1;33m\033[1;41m"),
    'DEBUG_FILTERS': COLOR_SEQ % "\033[0;35m",
}

DEBUG_FILTERS = 8
addLevelName(DEBUG_FILTERS, 'DEBUG_FILTERS')


# Global settings f logger.
settings = defaultdict(lambda: None)


def getLogger(name, parent=None):
    if isinstance(parent, LoggerAdapter):
        klass = type(parent)
        extra = parent.extra
        parent = parent.logger
    else:
        klass = None
        extra = None

    if parent:
        name = parent.name + '.' + name
    logger = _getLogger(name)
    logger.settings = settings

    if extra:
        logger = klass(logger, extra)
        logger.settings = settings
    return logger


class ColoredFormatter(Formatter):
    """
    Class written by airmind:
    https://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored
    """

    def format(self, record):
        levelname = record.levelname

        msg = super().format(record)
        if levelname in COLORS:
            msg = COLORS[levelname] % msg
        return msg


def createColoredFormatter(stream, format):
    if (sys.platform != 'win32') and stream.isatty():
        return ColoredFormatter(format)
    else:
        return Formatter(format)


if __name__ == '__main__':
    for levelname, cs in COLORS.items():
        print(cs % levelname, end=' ')



# Copyright(C) 2010-2014  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.base import empty
from woob.capabilities.weather import CapWeather
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter


__all__ = ['AppWeather']


class ForecastsFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'low', 'high')

    @staticmethod
    def temperature_display(t):
        return str(t.value)

    def format_obj(self, obj, alias):
        result = (
            '%s* %-15s%s (%s - %s)' % (
                self.BOLD,
                '%s:' % obj.date,
                self.NC,
                self.temperature_display(obj.low) if not empty(obj.low) else '?',
                self.temperature_display(obj.high) if not empty(obj.high) else '?'
            )
        )
        if hasattr(obj, 'text') and obj.text:
            result += ' %s' % obj.text
        return result


class CurrentFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'temp')

    temperature_display = staticmethod(lambda t: '%s' % t.value)

    def format_obj(self, obj, alias):
        result = '%s%s%s: %s' % (self.BOLD, obj.date, self.NC, self.temperature_display(obj.temp))
        if hasattr(obj, 'text') and obj.text:
            result += ' - %s' % obj.text
        return result


class CitiesFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name')

    def get_title(self, obj):
        return obj.name


class AppWeather(ReplApplication):
    APPNAME = 'weather'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to display weather and forecasts in your city."
    SHORT_DESCRIPTION = "display weather and forecasts"
    CAPS = CapWeather
    DEFAULT_FORMATTER = 'table'
    EXTRA_FORMATTERS = {'cities':    CitiesFormatter,
                        'current':   CurrentFormatter,
                        'forecasts': ForecastsFormatter,
                       }
    COMMANDS_FORMATTERS = {'cities':    'cities',
                           'current':   'current',
                           'forecasts': 'forecasts',
                          }

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    @defaultcount(10)
    def do_cities(self, pattern):
        """
        cities PATTERN

        Search cities.
        """
        self.change_path(['cities'])
        self.start_format()
        for city in self.do('iter_city_search', pattern, caps=CapWeather):
            self.cached_format(city)

    def complete_current(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_current(self, line):
        """
        current CITY_ID

        Get current weather for specified city. Use the 'cities' command to find them.
        """
        city, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(city)

        tr = self.config.get('settings', 'temperature_display', default='C')
        if tr == 'C':
            self.formatter.temperature_display = lambda t: t.ascelsius()
        elif tr == 'F':
            self.formatter.temperature_display = lambda t: t.asfahrenheit()

        self.start_format()
        for current in self.do('get_current', _id, backends=backend_name, caps=CapWeather):
            if current:
                self.format(current)

    def complete_forecasts(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_forecasts(self, line):
        """
        forecasts CITY_ID

        Get forecasts for specified city. Use the 'cities' command to find them.
        """
        city, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(city)

        tr = self.config.get('settings', 'temperature_display', default='C')
        if tr == 'C':
            self.formatter.temperature_display = lambda t: t.ascelsius()
        elif tr == 'F':
            self.formatter.temperature_display = lambda t: t.asfahrenheit()
        self.start_format()

        for forecast in self.do('iter_forecast', _id, backends=backend_name, caps=CapWeather):
            self.format(forecast)



# Copyright(C) 2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import datetime
from decimal import Decimal
from html import unescape
from urllib.parse import urljoin

import lxml.html as html

from woob.tools.html import html2text

from .base import (
    _NO_DEFAULT, Filter, FilterError, _Selector, debug, ItemNotFound,
    _Filter,
)
from .standard import CleanText

__all__ = ['CSS', 'XPath', 'XPathNotFound', 'AttributeNotFound',
           'Attr', 'Link', 'AbsoluteLink',
           'CleanHTML', 'FormValue', 'HasElement',
           'TableCell', 'ColumnNotFound',
           'ReplaceEntities',
          ]


class XPathNotFound(ItemNotFound):
    pass


class AttributeNotFound(ItemNotFound):
    pass


class ColumnNotFound(FilterError):
    pass


class CSS(_Selector):
    """Select HTML elements with a CSS selector

    For example::

        obj_foo = CleanText(CSS('div.main'))

    will take the text of all ``<div>`` having CSS class "main".
    """
    def select(self, selector, item):
        ret = item.cssselect(selector)
        if isinstance(ret, list):
            for el in ret:
                if isinstance(el, html.HtmlElement):
                    self.highlight_el(el, item)

        return ret


class XPath(_Selector):
    """Select HTML elements with a XPath selector
    """
    pass


class Attr(Filter):
    """Get the text value of an HTML attribute.

    Get value from attribute `attr` of HTML element matched by `selector`.

    For example::

        obj_foo = Attr('//img[@id="thumbnail"]', 'src')

    will take the "src" attribute of ``<img>`` whose "id" is "thumbnail".
    """

    def __init__(self, selector, attr, default=_NO_DEFAULT):
        """
        :param selector: selector targeting the element
        :param attr: name of the attribute to take
        """

        super(Attr, self).__init__(selector, default=default)
        self.attr = attr

    @debug()
    def filter(self, el):
        """
        :raises: :class:`XPathNotFound` if no element is found
        :raises: :class:`AttributeNotFound` if the element doesn't have the requested attribute
        """

        try:
            return '%s' % el[0].attrib[self.attr]
        except IndexError:
            return self.default_or_raise(XPathNotFound('Unable to find element %s' % self.selector))
        except KeyError:
            return self.default_or_raise(AttributeNotFound('Element %s does not have attribute %s' % (el[0], self.attr)))


class Link(Attr):
    """
    Get the link uri of an element.

    If the ``<a>`` tag is not found, an exception `IndexError` is raised.
    """

    def __init__(self, selector=None, default=_NO_DEFAULT):
        super(Link, self).__init__(selector, 'href', default=default)


class AbsoluteLink(Link):
    """Get the absolute link URI of an element.
    """
    def __call__(self, item):
        ret = super(AbsoluteLink, self).__call__(item)
        if ret:
            ret = urljoin(item.page.url, ret)
        return ret


class CleanHTML(Filter):
    """Convert HTML to text (Markdown) using html2text.

    .. seealso:: `html2text site <https://pypi.python.org/pypi/html2text>`_
    """

    def __init__(self, selector=None, options=None, default=_NO_DEFAULT):
        """
        :param options: options suitable for html2text
        :type options: dict
        """

        super(CleanHTML, self).__init__(selector=selector, default=default)
        self.options = options

    @debug()
    def filter(self, txt):
        if isinstance(txt, (tuple, list)):
            return ' '.join([self.clean(item, self.options) for item in txt])
        return self.clean(txt, self.options)

    @classmethod
    def clean(cls, txt, options=None):
        if not isinstance(txt, str):
            txt = html.tostring(txt, encoding="unicode")
        options = options or {}
        return html2text(txt, **options)


class UnrecognizedElement(Exception):
    pass


class FormValue(Filter):
    """
    Extract a Python value from a form element.

    Checkboxes and radio return booleans, while the rest
    return text. For ``<select>`` tags, returns the user-visible text.
    """

    @debug()
    def filter(self, el):
        try:
            el = el[0]
        except IndexError:
            return self.default_or_raise(XPathNotFound('Unable to find element %s' % self.selector))
        if el.tag == 'input':
            # checkboxes or radios
            if el.attrib.get('type') in ('radio', 'checkbox'):
                return 'checked' in el.attrib
            # regular text input
            elif el.attrib.get('type', '') in ('', 'text', 'email', 'search', 'tel', 'url', 'password', 'hidden', 'color'):
                try:
                    return str(el.attrib['value'])
                except KeyError:
                    return self.default_or_raise(AttributeNotFound('Element %s does not have attribute value' % el))
            # numeric input
            elif el.attrib.get('type', '') in ('number', 'range'):
                try:
                    if '.' in el.attrib.get('step', ''):

                        return Decimal(el.attrib['value'])
                    else:
                        return int(el.attrib['value'])
                except KeyError:
                    return self.default_or_raise(AttributeNotFound('Element %s does not have attribute value' % el))
            # datetime input
            try:
                if el.attrib.get('type', '') == 'date':
                    return datetime.datetime.strptime(el.attrib['value'], '%Y-%m-%d').date()
                elif el.attrib.get('type', '') == 'time':
                    return datetime.datetime.strptime(el.attrib['value'], '%H:%M').time()
                elif el.attrib.get('type', '') == 'datetime-local':
                    return datetime.datetime.strptime(el.attrib['value'], '%Y-%m-%dT%H:%M')
            except KeyError:
                return self.default_or_raise(AttributeNotFound('Element %s does not have attribute value' % el))
            else:
                raise UnrecognizedElement('Element %s is not recognized' % el)
        elif el.tag == 'textarea':
            return str(el.text)
        elif el.tag == 'select':
            options = el.xpath('.//option[@selected]')
            # default is the first one
            if len(options) == 0:
                options = el.xpath('.//option[1]')
            return '\n'.join(str(o.text) for o in options)
        else:
            raise UnrecognizedElement('Element %s is not recognized' % el)


class HasElement(Filter):
    """
    Returns `yesvalue` if the `selector` finds elements, `novalue` otherwise.
    """
    def __init__(self, selector, yesvalue=True, novalue=False):
        super(HasElement, self).__init__(selector, default=novalue)
        self.yesvalue = yesvalue

    @debug()
    def filter(self, value):
        if value:
            return self.yesvalue
        return self.default_or_raise(FilterError('No default value'))


class ReplaceEntities(CleanText):
    """
    Filter to replace HTML entities like "&eacute;" or "&#x42;" with their unicode counterpart.
    """
    def filter(self, data):
        txt = super(ReplaceEntities, self).filter(data)
        return unescape(txt)


class TableCell(_Filter):
    """
    Used with TableElement, gets the cell element from its name.

    For example:

    >>> from woob.capabilities.bank import Transaction
    >>> from woob.browser.elements import TableElement, ItemElement
    >>> class table(TableElement):
    ...     head_xpath = '//table/thead/th'
    ...     item_xpath = '//table/tbody/tr'
    ...     col_date =    u'Date'
    ...     col_label =   [u'Name', u'Label']
    ...     class item(ItemElement):
    ...         klass = Transaction
    ...         obj_date = Date(TableCell('date'))
    ...         obj_label = CleanText(TableCell('label'))
    ...

    TableCell handles table tags that have
    a "colspan" attribute that modify the width of the column:
    for example <td colspan="2"> will occupy two columns instead of one,
    creating a column shift for all the next columns that must be taken
    in consideration when trying to match columns values with column heads.
    """

    def __init__(self, *names, **kwargs):
        support_th = kwargs.pop('support_th', False)
        kwargs.pop('colspan', True)
        super(TableCell, self).__init__(**kwargs)
        self.names = names

        if support_th:
            self.td = '(./th | ./td)[%s]'
        else:
            self.td = './td[%s]'

    def __call__(self, item):
        # New behavior, handling colspans > 1
        for name in self.names:
            col_idx = item.parent.get_colnum(name)
            if col_idx is not None:
                current_col = 0
                for td_idx in range(col_idx + 1):
                    ret = item.xpath(self.td % (td_idx + 1))
                    if col_idx <= current_col:
                        for el in ret:
                            self.highlight_el(el, item)
                        return ret

                    if not ret:
                        # There might no be no TD at all
                        # ColumnNotFound seems for case when corresponding header is not found
                        # Thus for compat return empty
                        return []

                    current_col += int(ret[0].attrib.get('colspan', 1))

        return self.default_or_raise(ColumnNotFound('Unable to find column %s' % ' or '.join(self.names)))



# Copyright(C) 2014 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from functools import wraps
import re
from typing import Callable, Dict, Optional, TYPE_CHECKING, Tuple, Type, TypeVar
from urllib.parse import unquote

import requests

from woob.browser.pages import Page
from woob.browser.filters.base import _Filter
from woob.tools.regex_helper import normalize

if TYPE_CHECKING:
    from woob.browser.browsers import Browser

ABSOLUTE_URL_PATTERN_RE = re.compile(r'^[\w\?]+://[^/].*')

URLType = TypeVar('URLType', bound='URL')


class UrlNotResolvable(Exception):
    """
    Raised when trying to locate on an URL instance which url pattern is not resolvable as a real url.
    """


class URL:
    """
    A description of an URL on the PagesBrowser website.

    It takes one or several regexps to match urls, and an optional Page
    class which is instancied by PagesBrowser.open if the page matches a regex.

    .. warning::

        The ``methods`` parameter is only used for page matching, not request
        building using :py:meth:`URL.go` or :py:meth:`URL.open`; you must
        still set the method using these.

    :param base: The name of the browser's property containing the base URL.
    :param headers: Headers to include on requests using this URL.
    :param timeout: Timeout to use for this URL in particular.
    :param methods: Request HTTP methods to match the response.
    :param content_type: MIME type of the content to match the response with.
    """
    _creation_counter = 0

    def __init__(
        self, *args,
        base: str = 'BASEURL',
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        methods: Tuple[str, ...] = (),
        content_type: Optional[str] = None,
    ):
        if content_type is not None and ';' in content_type:
            raise ValueError(
                'Content-Type matching is only based on the MIME type, '
                + 'not additional properties such as encoding or version',
            )

        self.urls = []
        self.klass = None
        self.browser = None
        for arg in args:
            if isinstance(arg, str):
                self.urls.append(arg)
            if isinstance(arg, type):
                self.klass = arg

        self._base = base
        self._headers = headers
        self._timeout = timeout
        self._methods = tuple(methods)
        self._content_type = content_type
        self._creation_counter = URL._creation_counter
        URL._creation_counter += 1

    def is_here(self, **kwargs) -> bool:
        """
        Returns True if the current page of browser matches this URL.
        If arguments are provided, and only then, they are checked against the arguments
        that were used to build the current page URL.
        """
        assert self.klass is not None, "You can use this method only if there is a Page class handler."
        assert self.browser is not None

        if self.browser.page is None:
            return False

        if len(kwargs):
            m = self.match(self.build(**kwargs))
            assert m is not None
            params = m.groupdict()
        else:
            params = None

        if not isinstance(self.browser.page, self.klass):
            return False

        if self._methods:
            method = self.browser.response.request.method
            if method not in self._methods:
                return False

        if self._content_type is not None:
            content_type = self.browser.response.headers.get('Content-Type')
            if content_type is None:
                return False

            content_type, _, _ = content_type.partition(';')
            content_type = content_type.strip()
            if content_type != self._content_type:
                return False

        # XXX use unquote on current params values because if there are spaces
        # or special characters in them, it is encoded only in but not in kwargs.
        return (
            params is None or
            params == {k: unquote(v) for k, v in self.browser.page.params.items()}
        )

    def stay_or_go(
        self,
        params: Dict | None = None,
        data: str | Dict | None = None,
        json: Dict | None = None,
        method: str | None = None,
        headers: Dict[str, str] | None = None,
        **kwargs
    ) -> requests.Response | Page:
        """
        Request to go on this url only if we aren't already here.

        Arguments are optional parameters for url.

        >>> url = URL('https://exawple.org/(?P<pagename>).html')
        >>> url.stay_or_go(pagename='index')
        """
        assert self.browser is not None

        if self.is_here(**kwargs):
            return self.browser.page

        return self.go(params=params, data=data, json=json, method=method, headers=headers, **kwargs)

    def go(
        self,
        *,
        params: Dict | None = None,
        data: str | Dict | None = None,
        json: Dict | None = None,
        method: str | None = None,
        headers: Dict[str, str] |  None = None,
        timeout: Optional[float] = None,
        **kwargs
    ) -> requests.Response | Page:
        """
        Request to go on this url.

        Arguments are optional parameters for url.

        >>> url = URL('https://exawple.org/(?P<pagename>).html')
        >>> url.stay_or_go(pagename='index')
        """
        assert self.browser is not None

        headers = headers or {}
        if self._headers:
            headers.update(self._headers)

        if timeout is None:
            timeout = self._timeout

        r = self.browser.location(
            self.build(**kwargs),
            params=params,
            data=data,
            json=json,
            method=method,
            headers=headers,
            timeout=timeout,
        )
        return r.page or r

    def open(
        self,
        *,
        params: Dict | None = None,
        data: Dict | str | None = None,
        json: Dict | None = None,
        method: str | None = None,
        headers: Dict[str, str] | None = None,
        timeout: float | None = None,
        is_async: bool = False,
        callback: Callable[[requests.Response], requests.Response] = lambda response: response,
        **kwargs
    ) -> requests.Response | Page:
        """
        Request to open on this url.

        Arguments are optional parameters for url.

        >>> url = URL('https://exawple.org/(?P<pagename>).html')
        >>> url.open(pagename='index')
        """
        assert self.browser is not None

        headers = headers or {}
        if self._headers:
            headers.update(self._headers)

        if timeout is not None:
            timeout = self._timeout

        r = self.browser.open(
            self.build(**kwargs),
            params=params,
            data=data,
            json=json,
            method=method,
            headers=headers,
            is_async=is_async,
            callback=callback,
        )

        if hasattr(r, 'page') and r.page:
            return r.page
        return r

    def get_base_url(
        self,
        browser: Browser | None = None,
        for_pattern: str | None = None
    ) -> str:
        """
        Get the browser's base URL for the instance.

        ``for_pattern`` argument is optional and only used to display more
        information in the ValueError exception (don't know why, may be
        removed).
        """
        browser = browser or self.browser
        if browser is None:
            raise ValueError('URL browser is not set')

        value = getattr(browser, self._base, None)
        if not isinstance(value, str):
            msg = f'Browser {self._base} property is None or not defined'
            if for_pattern:
                msg += f', URL {for_pattern} should be defined as absolute'
            raise ValueError(msg)

        return value

    def build(self, **kwargs) -> str:
        """
        Build an url with the given arguments from URL's regexps.

        :param param: Query string parameters

        :rtype: :class:`str`
        :raises: :class:`UrlNotResolvable` if unable to resolve a correct url with the given arguments.
        """
        browser = kwargs.pop('browser', self.browser)

        assert browser is not None

        params = kwargs.pop('params', None)
        patterns = []
        for url in self.urls:
            patterns += normalize(url)

        base = None
        for pattern, _ in patterns:
            url = pattern
            # only use full-name substitutions, to allow % in URLs
            args = kwargs.copy()
            for key in list(args.keys()):  # need to use keys() because of pop()
                search = f'%({key})s'
                if search in pattern:
                    url = url.replace(search, str(args.pop(key)))
            # if there are named substitutions left, ignore pattern
            if re.search(r'%\([A-z_]+\)s', url):
                continue
            # if not all args were used
            if len(args):
                continue

            if base is None and not ABSOLUTE_URL_PATTERN_RE.match(url):
                base = self.get_base_url(browser=browser, for_pattern=url)
                url = browser.absurl(url, base=base)

            if params:
                p = requests.models.PreparedRequest()
                p.prepare_url(url, params)
                assert p.url is not None
                url = p.url
            return url

        raise UrlNotResolvable('Unable to resolve URL with %r. Available are %s' % (kwargs, ', '.join([pattern for pattern, _ in patterns])))

    def match(
        self, url: str,
        base: str | None = None
    ) -> re.Match | None:
        """
        Check if the given url match this object.

        Returns ``None`` if none matches.
        """
        for regex in self.urls:
            if not ABSOLUTE_URL_PATTERN_RE.match(regex):
                if not base:
                    base = self.get_base_url(browser=None, for_pattern=regex)

                regex = re.escape(base).rstrip('/') + '/' + regex.lstrip('/')

            m = re.match(regex, url)
            if m:
                return m

        return None

    def handle(self, response: requests.Response) -> Page | None:
        """
        Handle a HTTP response to get an instance of the klass if it matches.
        """
        assert self.browser is not None

        if self.klass is None:
            return None
        if response.request.method == 'HEAD':
            return None
        if self._methods and response.request.method not in self._methods:
            return None
        if self._content_type is not None:
            content_type = response.headers.get('Content-Type')
            if content_type is None:
                return None

            content_type, _, _ = content_type.partition(';')
            content_type = content_type.strip()
            if content_type != self._content_type:
                return None

        m = self.match(response.url)
        if m:
            page = self.klass(self.browser, response, m.groupdict())
            if hasattr(page, 'is_here'):
                if page.is_here is None or page.is_here is True:
                    return page
                elif page.is_here is False:
                    return None  # no page!
                elif isinstance(page.is_here, _Filter):
                    if page.is_here(page.doc):
                        return page
                elif callable(page.is_here):
                    if page.is_here():
                        return page
                else:
                    assert isinstance(page.is_here, str)
                    if page.doc.xpath(page.is_here):
                        return page
            else:
                return page

        return None

    def id2url(self, func: Callable):
        r"""
        Helper decorator to get an URL if the given first parameter is an ID.
        """

        @wraps(func)
        def inner(browser, id_or_url: str, *args, **kwargs):
            if re.match('^https?://.*', id_or_url):
                base = self.get_base_url(browser=browser)
                if not self.match(id_or_url, base=base):
                    return None
            else:
                id_or_url = self.build(id=id_or_url, browser=browser)

            return func(browser, id_or_url, *args, **kwargs)
        return inner

    def with_headers(
        self: URLType,
        headers: Optional[Dict[str, str]],
    ) -> URLType:
        """Get the current URL with different stored headers.

        For example, suppose that a browser needs to add an 'Accept'
        header for accessing a specific header of the API; see `Using the
        Accept Header to version your API`_ for more details.

        .. code-block:: python

            class MyBrowser(PagesBrowser):
                products = URL('products')

            class MyChildBrowser(MyBrowser):
                BASEURL = 'https://products-api.example/'

                products = MyBrowser.products.with_headers({
                    'Accept': 'application/vnd.example.api+json;version=2',
                })

        .. _`Using the Accept Header to version your API`:
            https://labs.qandidate.com/blog/2014/10/16/using-the-accept-header
            -to-version-your-api/

        :param headers: The new headers to set to the URL.
        :return: The URL using the different headers.
        """
        new_url = self.__class__(
            *self.urls,
            self.klass,
            base=self._base,
            headers=headers,
            timeout=self._timeout,
            methods=self._methods,
            content_type=self._content_type,
        )
        new_url.browser = None
        return new_url

    def without_headers(self: URLType) -> URLType:
        """Get the current URL without stored headers.

        :return: The URL using the different headers.
        """
        return self.with_headers(None)

    def with_timeout(self: URLType, timeout: Optional[float]) -> URLType:
        """Get a new URL object with timeout.

        :param timeout: The new timeout to apply, or ``None`` if the default
            timeout from the browser is to be used.
        :return: The URL using the different timeout.
        """
        new_url = self.__class__(
            *self.urls,
            self.klass,
            base=self._base,
            headers=self._headers,
            timeout=timeout,
            methods=self._methods,
            content_type=self._content_type,
        )
        new_url.browser = None
        return new_url

    def without_timeout(self: URLType) -> URLType:
        """Get a new URL object using the browser's timeout.

        :return: The URL without the custom timeout.
        """
        return self.with_timeout(None)

    def with_page(self: URLType, cls: Type[Page]) -> URLType:
        """Get a new URL with the same path but a different page class.

        :param cls: The new page class to use.
        :return: The URL object with the updated page class.
        """
        new_url = self.__class__(
            *self.urls,
            cls,
            base=self._base,
            headers=self._headers,
            timeout=self._timeout,
            methods=self._methods,
            content_type=self._content_type,
        )
        new_url.browser = None
        return new_url

    def with_urls(
        self: URLType,
        *urls: str,
        clear: bool = True,
        match_new_first: bool = True
    ) -> URLType:
        """Get a new URL object with the same page but with different paths.

        :param urls: List of urls handled by the page.
        :param clear: If True, the page will only handled the given urls.
            Otherwise, the urls are added to already handled urls.
        :param match_new_first: If true, new paths will be matched first
            for this URL; this parameter is ignored when ``clear`` is True.
        :return: The URL object with the updated patterns.
        """
        new_urls = list(urls)
        if not clear:
            # needed to extend self.urls which is a list
            if match_new_first:
                new_urls = new_urls + self.urls
            else:
                new_urls = self.urls + new_urls

        # We only want unique patterns here.
        new_urls = list(dict.fromkeys(new_urls))

        new_url = self.__class__(
            *new_urls,
            self.klass,
            base=self._base,
            headers=self._headers,
            timeout=self._timeout,
            methods=self._methods,
            content_type=self._content_type,
        )
        new_url.browser = None
        return new_url

    def with_base(
        self: URLType,
        base: str = 'BASEURL',
    ) -> URLType:
        """Get a new URL object with a custom base.

        :param base: The name of the new base, or None to use the default one.
        :return: The URL object with the updated base.
        """
        return self.__class__(
            *self.urls,
            self.klass,
            base=base,
            headers=self._headers,
            timeout=self._timeout,
            methods=self._methods,
            content_type=self._content_type,
        )

    def with_methods(
        self: URLType,
        methods: Tuple[str, ...],
    ) -> URLType:
        """Get a new URL object with custom methods.

        :param methods: The new methods to match the URL with.
        :return: The URL object with the updated methods.
        """
        return self.__class__(
            *self.urls,
            self.klass,
            base=self._base,
            headers=self._headers,
            timeout=self._timeout,
            methods=methods,
            content_type=self._content_type,
        )

    def without_methods(self: URLType) -> URLType:
        """Get a new URL object without matching on methods.

        :return: The URL object with the updated methods.
        """
        return self.with_methods(())

    def with_content_type(self: URLType, content_type: Optional[str]) -> URLType:
        """Get a new URL object with custom Content-Type matching.

        :param content_type: The new content type to match with.
        :return: The URL object with the updated content type to match.
        """
        return self.__class__(
            *self.urls,
            self.klass,
            base=self._base,
            headers=self._headers,
            timeout=self._timeout,
            methods=self._methods,
            content_type=content_type,
        )

    def without_content_type(self: URLType) -> URLType:
        """Get a new URL object without Content-Type matching.

        :return: The URL object with no content type matching.
        """
        return self.with_content_type(None)


class BrowserParamURL(URL):
    r"""A URL that automatically fills some params from browser attributes.

    URL patterns having groups named "browser_*" will pick the relevant
    attribute from the browser. For example:

        foo = BrowserParamURL(r'/foo\?bar=(?P<browser_token>\w+)')

    The browser is expected to have a `.token` attribute and it will be passed
    automatically when just calling `foo.go()`, it's equivalent to
    `foo.go(browser_token=browser.token)`.

    Warning: all `browser_*` params will be passed, having multiple patterns
    with different groups in a `BrowserParamURL` is risky.
    """

    def build(self, **kwargs) -> str:
        prefix = 'browser_'

        for arg in kwargs:
            if arg.startswith(prefix):
                raise ValueError('parameter %r is reserved by URL pattern')

        for url in self.urls:
            for groupname in re.compile(url).groupindex:
                if groupname.startswith(prefix):
                    attrname = groupname[len(prefix):]
                    kwargs[groupname] = getattr(self.browser, attrname)

        return super().build(**kwargs)


def normalize_url(url: str) -> str:
    """Normalize URL by lower-casing the domain and other fixes.

    Lower-cases the domain, removes the default port and a trailing dot.

    >>> normalize_url('https://EXAMPLE:80')
    'https://example'
    """

    def norm_domain(m):
        # don't use urlparse/urlunparse because it might do too much normalization

        auth, authsep, hostport = m.group(2).rpartition('@')
        host, portsep, port = hostport.partition(':')

        if (
            (port == '443' and m.group(1) == 'https://')
            or (port == '80' and m.group(1) == 'http://')
        ):
            portsep = port = ''

        host = host.lower().rstrip('.')

        return ''.join((m.group(1), auth, authsep, host, portsep, port))

    return re.sub(r'^(https?://)([^/#?]+)', norm_domain, url)



# Copyright(C) 2010-2013 Romain Bignon, Christophe Benz
# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .date import DeltaField
from .image import CapImage, BaseImage


__all__ = ['BaseVideo', 'CapVideo']


class BaseVideo(BaseImage):
    """
    Represents a video.

    This object has to be inherited to specify how to calculate the URL of the video from its ID.
    """
    duration =  DeltaField('file duration')


class CapVideo(CapImage):
    """
    Video file provider.
    """

    def search_videos(self, pattern, sortby=CapImage.SEARCH_RELEVANCE, nsfw=False):
        """
        search for a video file

        :param pattern: pattern to search on
        :type pattern: str
        :param sortby: sort by... (use SEARCH_* constants)
        :param nsfw: include non-suitable for work videos if True
        :type nsfw: bool
        :rtype: iter[:class:`BaseVideo`]
        """
        return self.search_image(pattern, sortby, nsfw)

    def get_video(self, _id):
        """
        Get a video file from an ID.

        :param _id: video file ID
        :type _id: str
        :rtype: :class:`BaseVideo` or None is fot found.
        """
        return self.get_image(_id)



# Copyright(C) 2009-2015  Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.browser.filters.standard import _Filter, Field, debug
from woob.capabilities.base import empty
from decimal import Decimal


class PricePerMeterFilter(_Filter):
    """
    Filter that help to fill PricePerMeter field
    """
    def __init__(self):
        super(PricePerMeterFilter, self).__init__()

    @debug()
    def __call__(self, item):
        cost = Field('cost')(item)
        area = Field('area')(item)
        if not (empty(cost) or empty(area)):
            return Decimal(cost or 0) / Decimal(area or 1)
        return Decimal(0)



# Copyright(C) 2010-2013 Christophe Benz, Romain Bignon, Julien Hebert
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import datetime

from woob.tools.date import new_date, new_datetime
from woob.capabilities.base import Field


__all__ = ['DateField', 'TimeField', 'DeltaField']


class DateField(Field):
    """
    A field which accepts only :class:`datetime.date` and :class:`datetime.datetime` types.
    """

    def __init__(self, doc, **kwargs):
        super(DateField, self).__init__(doc, datetime.date, datetime.datetime, **kwargs)

    def __setattr__(self, name, value):
        if name == 'value':
            # Force use of our date and datetime types, to fix bugs in python2
            # with strftime on year<1900.
            if type(value) is datetime.datetime:
                value = new_datetime(value)
            if type(value) is datetime.date:
                value = new_date(value)
        return object.__setattr__(self, name, value)


class TimeField(Field):
    """
    A field which accepts only :class:`datetime.time` and :class:`datetime.time` types.
    """

    def __init__(self, doc, **kwargs):
        super(TimeField, self).__init__(doc, datetime.time, datetime.datetime, **kwargs)


class DeltaField(Field):
    """
    A field which accepts only :class:`datetime.timedelta` type.
    """

    def __init__(self, doc, **kwargs):
        super(DeltaField, self).__init__(doc, datetime.timedelta, **kwargs)

    def convert(self, value):
        if isinstance(value, int):
            value = datetime.timedelta(seconds=value)
        return value



# Copyright(C) 2014 Oleg Plakhotniuk
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import re

__all__ = ['ReTokenizer']


class ReTokenizer:
    """
    Simple regex-based tokenizer (AKA lexer or lexical analyser).
    Useful for PDF statements parsing.

    1. There's a lexing table consisting of type-regex tuples.
    2. Lexer splits text into chunks using the separator character.
    3. Text chunk is sequentially matched against regexes and first
       successful match defines the type of the token.

    Check out test() function below for examples.
    """

    def __init__(self, text, sep, lex):
        self._lex = lex
        self._tok = [ReToken(lex, chunk) for chunk in text.split(sep)]

    def tok(self, index):
        if 0 <= index < len(self._tok):
            return self._tok[index]
        else:
            return ReToken(self._lex, eof=True)

    def simple_read(self, token_type, pos, transform=lambda v: v):
        t = self.tok(pos)
        is_type = getattr(t, 'is_%s' % token_type)()
        return (pos+1, transform(t.value())) if is_type else (pos, None)


class ReToken:
    def __init__(self, lex, chunk=None, eof=False):
        self._lex = lex
        self._eof = eof
        self._value = None
        self._type = None
        if chunk is not None:
            for type_, regex in self._lex:
                m = re.match(regex, chunk, flags=re.UNICODE)
                if m:
                    self._type = type_
                    if len(m.groups()) == 1:
                        self._value = m.groups()[0]
                    elif m.groups():
                        self._value = m.groups()
                    else:
                        self._value = m.group(0)
                    break

    def is_eof(self):
        return self._eof

    def value(self):
        return self._value

    def __getattr__(self, name):
        if name.startswith('is_'):
            return lambda: self._type == name[3:]
        raise AttributeError()



# Copyright(C) 2017  Vincent A
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.
#
# This file is distributed under the WTFPLv2 license.

import sys
import os

import PIL.Image as Image
import PIL.ImageColor as PILColor

"""
XTerm can decode sequences and display 256 colors:
- 16 system colors, supported by many terms [0-15]
- 216 colors (not the web-safe palette) [16-231]
- 24 grey colors (excluding black and white) [231-255]

Other terminals do support these 256 colors escape codes, e.g. roxterm, xfce-terminal.
"""

__all__ = ('make256xterm', 'imageRGB_to_256', 'image2xterm', 'image256_to_ansi')


def make256xterm():
    """Return a [r, g, b, r, g, b, ...] table with the colors of the 256 xterm colors.

    The table is indexed: [0:3] corresponds to color 0.
    """
    color256 = []
    # standard 16 colors
    color256 += list(sum((PILColor.getrgb(c) for c in
        '''black maroon green olive navy purple teal silver
        gray red lime yellow blue fuchsia aqua white'''.split()), ()))

    steps = (0, 95, 135, 175, 215, 255)
    for r in steps:
        for g in steps:
            for b in steps:
                color256 += [r, g, b]

    # grays
    for v in range(8, 248, 10):
        color256 += [v, v, v]

    assert len(color256) == 256 * 3

    return color256


TABLE_XTERM_256 = make256xterm()


def imageRGB_to_256(im):
    """Returns `im` converted to the XTerm 256 colors palette.

    The image should be resized *before* applying this function.
    """
    paletter = Image.new('P', (1, 1))
    paletter.putpalette(TABLE_XTERM_256)
    return im.quantize(palette=paletter)


def image256_to_ansi(im, halfblocks):
    """Print PIL image `im` to `fd` using XTerm escape codes.

    1 pixel corresponds to exactly 1 character on the terminal. 1 row of pixels is terminated by a newline.
    `im` has to be a 256 colors image (ideally in the XTerm palette if you want it to make sense).
    `im` should be resized appropriately to fit terminal size and character aspect (characters are never square).
    """
    buf = []
    pix = im.load()
    w, h = im.size

    if halfblocks:
        yrange = range(0, h, 2)
    else:
        yrange = range(h)

    for y in yrange:
        for x in range(w):
            if halfblocks:
                if y + 1 >= h:
                    buf.append('\x1b[38;5;%dm\u2580' % pix[x, y])
                else:
                    buf.append('\x1b[38;5;%dm\x1b[48;5;%dm\u2580' % (pix[x, y], pix[x, y + 1]))
            else:
                buf.append('\x1b[48;5;%dm ' % pix[x, y])
        buf.append('\x1b[0m%s' % os.linesep)
    return ''.join(buf)


def image2xterm(imagepath, newsize=(80, 24), halfblocks=True):
    image = Image.open(imagepath)
    image.load()

    stretch = 2
    curratio = image.size[1] / (image.size[0] * stretch)
    targetsize = newsize[0], int(newsize[0] * curratio)
    if targetsize[1] > newsize[1]:
        targetsize = int(newsize[1] / curratio), newsize[1]
    if halfblocks:
        targetsize = targetsize[0], targetsize[1] * 2

    image = image.convert('RGB')
    image = image.resize(targetsize, Image.ANTIALIAS)
    image2 = imageRGB_to_256(image)
    return image256_to_ansi(image2, halfblocks=halfblocks)


def _getoutsize(infoname, default):
    """Get suitable dimension for tty or default"""
    if sys.stdout.isatty():
        import curses
        curses.setupterm()
        return curses.tigetnum(infoname)
    else:
        return default


def get_term_size():
    return (_getoutsize('cols', 80), _getoutsize('lines', 24))


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Print an image on terminal with 256 colors palette')
    parser.add_argument('-r', '--rows', dest='rows', metavar='SIZE', type=int)
    parser.add_argument('-c', '--columns', dest='cols', metavar='SIZE', type=int)
    parser.add_argument('--spaces', action='store_true')
    parser.add_argument('file')
    opts = parser.parse_args()

    if opts.rows is None:
        opts.rows = _getoutsize('lines', 24)
    if opts.cols is None:
        opts.cols = _getoutsize('cols', 80)

    data = image2xterm(opts.file, (opts.cols, opts.rows), halfblocks=not opts.spaces)
    sys.stdout.write(data)


if __name__ == '__main__':
    main()
